/*
 * Created on Aug 13, 2004
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.bridge;


import java.math.BigInteger;
import java.util.*;

import org.msda.components.core.MSDDaemon;
import org.msda.components.interfaces.MSDABridgeInterface;
import org.msda.components.interfaces.MSDANetworkInterface;
import org.msda.components.utils.LocalComponentsMgr;
import org.msda.context.*;
import org.msda.framework.FrameworkOutput;
import org.msda.msd.*;
import org.msda.msd.impl.*;
import org.msda.util.*;
import org.msda.util.comm.NetworkUtils;


/**
 * This class implements the component that propagates meta-descriptions and meta-requests between different networks (MSDA Managers)
 * At first, the MSDA Bridge component is only given a name and the intereface to the core component. 
 * Later, specific network interfaces/multicast groups are added to the bridge that will be used to 
 * propagate the meta-descriptions and meta-requesta. Each time a new "interface" is added, the bridge
 * component has to contact the msd manager in charge of the network and register with it.
 * 
 * @author raverdy
 *
 */

public class MSDBridge implements MSDABridgeInterface {
    
	private Vector activeInterfacesVector;
	private MSDDaemon daemon;
	private MSDANetworkInterface networkInterface;
	private int bridgingInterfaceCounter;
	//	 Identification of the bridge 
	private UbisecID bridgeId;
	private String ipAddress;
	private int remoteAccessPort;
	private LocalComponentsMgr localMgr;
	private MSDBridgePool bridgePool;
	private FrameworkOutput board;
	
	// Needed for the gui
	private Vector networkInterfacesVector;
	
	
	/**
	 * 
	 * @param daemon
	 * @param networks
	 * @param name
	 * @param isContextEnabled
	 * @param board
	 * @throws MSDException
	 */
	public MSDBridge ( MSDDaemon daemon, List networks, String name ) throws MSDException
	{
		this.daemon = daemon;
		this.board = org.msda.framework.FrameworkMain.msdaOutput;
		this.networkInterfacesVector = new Vector();
		this.activeInterfacesVector = new Vector();
		this.bridgingInterfaceCounter = 0;
		//
		this.bridgeId = new UbisecIDImpl();
		this.bridgeId.setName( name );
		this.localMgr = new LocalComponentsMgr();
		this.localMgr.setBoard( board );
		this.bridgePool = new MSDBridgePool( localMgr, networks, daemon ); 
		//
		this.networkInterface = daemon.getNetworkInterface();
	}
	
	
	/**
	 * 
	 * @see java.lang.Object#finalize()
	 */
	public void finalize () throws Throwable
	{
		ThreadBridgingInterface tmpInterface;
		NetworkConfig networkConfig;
		try {
			bridgeId = null;
			ipAddress = null;
			bridgePool = null;
			localMgr = null;
			board = null;
			daemon = null;
			networkInterface = null;
			
			while (activeInterfacesVector.size()>0) {
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.remove(activeInterfacesVector.size()-1);
				tmpInterface.terminate();
				tmpInterface = null;
			}
			
			while (networkInterfacesVector.size()>0) {
				networkConfig = (NetworkConfig)networkInterfacesVector.remove(networkInterfacesVector.size()-1);
				networkConfig = null;
			}
		} catch (Exception e) {
		} finally {
			activeInterfacesVector = null;
			networkInterfacesVector = null;
			super.finalize ();
		}
	}
	
	/**
	 * Returns the DashBoard.
	 * 
	 * @return Returns the board.
	 */
	public FrameworkOutput getBoard() {
		return board;
	}
	
	/**
	 * Returns the LocalComponentsMgr.
	 * 
	 * @return Returns the localMgr.
	 */
	public LocalComponentsMgr getLocalMgr() {
		return localMgr;
	}
	
	/**
	 * Returns the MSDBridgePool.
	 * 
	 * @return Returns the bridgePool.
	 */
	public MSDBridgePool getBridgePool() {
		return bridgePool;
	}
	
	/**
	 * Returns the network interface.
	 * 
	 * @return Returns the network interface.
	 */
	public MSDANetworkInterface getNetworkInterface() {
		return networkInterface;
	}
	
	/**
	 * Returns the Vector of active interfaces
	 * 
	 * @return Returns the Vector of active interfaces
	 */
	public Vector getActiveInterfacesVector() {
		return activeInterfacesVector;
	}
	
	/**
	 * Returns the bridge UbisecID
	 * 
	 * @return Returns the bridge UbisecID
	 */
	public UbisecID getBridgeID() {
		return bridgeId;
	}
	
	/**
	 * Returns the IP address
	 * 
	 * @return Returns the IP address.
	 */
	public String getIpAddress() {
		return ipAddress;
	}
	
	/**
	 * Returns the remote access port.
	 * 
	 * @return Returns the remote access port.
	 */
	public int getRemoteAccessPort() {
		return remoteAccessPort;
	}
	
	/**
	 * Returns if the network interface identified by <code>networkConfig</code> is active or not. 
	 * 
	 * @param networkConfig	The network interface
	 * @return	true if the network interface is active, false otherwise
	 */
	public boolean isActiveInterface (NetworkConfig networkConfig)
	{
		ThreadBridgingInterface tmpInterface;
		
		//board.info("1)- check if bridging interface active " + networkConfig.getInterface() + " " + networkConfig.getDataGroup() + ":" + networkConfig.getDataPort().intValue() );		
		
		synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++) 
			{
				UbisecID tmpID =new UbisecIDImpl();
				tmpID.setName(networkConfig.getName());
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (MSDUtil.getInstance().compareID(tmpInterface.getNetId(),tmpID))
				{
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Creates an entry for a new network to connect to. <br>
	 * In theory, all manager/bridges communication in different networks should use the same default multicast group and 
	 * port but for testing reasons (not enough computers), the group and port can be specified.
	 * This method needs to be synchronized as it manipulates the interface list.
	 *
	 * @param networkConfig that contains <p>
	 *  		netname 	name of the network (for id purposes) <br>
	 * 			ifname		name of nthe network interface to use (eth0...) <br>
	 * 			groupAddress	multicast group address to use for communicating with the MSDA manager <br>
	 * 			groupPort 	port number to use for communicating with the msd manager 
	 *  
	 * @return true if the interface could be added, false otherwise
	 */  
	public boolean activateInterface( NetworkConfig networkConfig ) 
	{
		ThreadBridgingInterface tmpInterface;
		boolean res = true;
		
		//board.info("1)- Activate a bridging interface on " + networkConfig.getInterface() + " " + networkConfig.getDataGroup() + ":" + networkConfig.getDataPort().intValue() );		
		
		synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++) 
			{
				UbisecID tmpID =new UbisecIDImpl();
				tmpID.setName(networkConfig.getName());
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (MSDUtil.getInstance().compareID(tmpInterface.getNetId(),tmpID))
				{
					return true;
				}
			}
			// Must catch exception if the interface/group/port parameters are wrong		
			try {
				bridgingInterfaceCounter++;
				ThreadBridgingInterface newInterface = new ThreadBridgingInterface( this, networkConfig, NetworkUtils.getIPAddress(networkConfig.getInterface()), bridgingInterfaceCounter );
				newInterface.start();
				activeInterfacesVector.add(newInterface);
				networkInterfacesVector.add( networkConfig );
				
				// add a new thread to the pool
				// WARNING : not correct, should remove a thread if the bridgingInterface is desactivated
				this.bridgePool.createNewRemoteAccessThread();
				
				// send bridge beacon
				daemon.sendBeaconBridge ( networkConfig );
				
				board.info ("- Bridging interface started..." );		
			}
			catch (MSDException e)
			{
				board.notifyError("- bridging interface failed" + networkConfig.getInterface() + " " + networkConfig.getDataGroup() + ":" + networkConfig.getDataPort().intValue() );
				res = false;
			}
		}
		return res;
	}
	
	
	/**
     * Desactivates the network interface given in the <code>networkConfig</code>
     * 
     * @param networkConfig	The network interface
	 * @see org.msda.components.interfaces.MSDABridgeInterface#desactivateInterface(org.msda.msd.NetworkConfig)
	 */
	public boolean desactivateInterface( NetworkConfig networkConfig ) 
	{
		ThreadBridgingInterface tmpInterface;
		boolean res = true;
		
		//board.info("1)- Desactivate a bridging interface on " + networkConfig.getInterface() + " " + networkConfig.getDataGroup() + ":" + networkConfig.getDataPort().intValue() );		
		//		board.info("- desactivate a bridging interface on " + networkConfig.getInterface() + " " + networkConfig.getDataGroup() + ":" + networkConfig.getDataPort().intValue() );		
		
		synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++) 
			{
				UbisecID tmpID =new UbisecIDImpl();
				tmpID.setName(networkConfig.getName());
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (MSDUtil.getInstance().compareID(tmpInterface.getNetId(),tmpID))
				{
					tmpInterface.terminate();
					tmpInterface = null;
					activeInterfacesVector.removeElementAt(x);
					x--;
				}
			}
		}
		return true;
	}
	
	
	/**
	 * Propagates an MSDA description received on an interface onto the other interfaces.
	 * This method needs to be synchronized as it manipulates the interface list.
	 * 
	 * @param srcId the source network the msd description is coming from
	 * @param bridgeInterfaceId the id of the bridge interface
	 * @param description the MSDA description to propagate
	 */
	public void processIncomingMessage ( UbisecID srcId, int bridgeInterfaceId, MSDDescription description )
	{
		board.info ("- Received a description from interface on " + bridgeInterfaceId+" "+srcId.getName());
		
		// When receiving a message, the MSD Bridge first checks that it has not yet forwarded the same
		// message. If so, the message is just discarted.
		if ( description.getPropaInfo()!=null ) {
			List hops = description.getPropaInfo().getHops();
			if ( hops!=null ) {
				for (int x=0; x<hops.size();x++)
				{
					Hop tmpHop = (Hop)hops.get(x);
					if ( MSDUtil.getInstance().compareID (tmpHop.getBridgeID(), this.bridgeId) )
					{
						// already forwarded 
						board.info ("  - Already forwarded this description, dump it - "+description.getDescID().getName() );
						return;
					}
				}
			}
		}
		
		// First, check all networks this bridge is connected to, and propagate the message
		// (except the one the message is coming from).
		ThreadBridgingInterface tmpInterface;
		//synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++)
			{
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				
				if (tmpInterface.getBridgeInterfaceId() == bridgeInterfaceId)
				{
					board.info ("- Received a description from interface on " + tmpInterface.getIfname() + " " + tmpInterface.getGroupAddress() + ":" + tmpInterface.getGroupPort() +" - NO FORWARDING" );
				} else {
					// Before propagating the message, a new "hop" is added to the propagation list
					// Change the destination network in each hope for each interface (see note below)
					board.info ("- Forward the description on interface on " + tmpInterface.getIfname() + " " + tmpInterface.getGroupAddress() + ":" + tmpInterface.getGroupPort() );
					Hop myHop = new HopImpl();
					myHop.setBridgeID(bridgeId);
					myHop.setSrcNetID(srcId);
					myHop.setDestNetID(tmpInterface.getNetId());
					myHop.setSecurity("nosecurity");
					myHop.setTime(Calendar.getInstance());
					if (description.getPropaInfo()==null ) {
						PropagationInformation propagation = new PropagationInformationImpl();
						description.setPropaInfo(propagation);
					}
					// WARNING: here the size of the list is recorded, because the hop added may then be removed
					// This is because the bridge may send the message to more than one network so the latest 
					// hop (destination network) may be different
					int lastPos = description.getPropaInfo().getHops().size();
					
					// Check that the message has not yet been on the network (from another bridge)
					List hops = description.getPropaInfo().getHops();
					if ( hops!=null ) {
						Iterator it = hops.iterator();
						while( it.hasNext() ) {
							Hop tmpHop = (Hop)it.next();
							
							if ( MSDUtil.getInstance().compareID( tmpHop.getDestNetID(), tmpInterface.getNetId() ) 
									|| MSDUtil.getInstance().compareID( tmpHop.getSrcNetID(), tmpInterface.getNetId() )  ) {
								board.info(" - Received a MSDDescription that has already been on this network..."); 
								return;   
							}
						}
					}
					
					// Add the hop and send the description
					description.getPropaInfo().getHops().add ( myHop );
					
					// The number of hops is ok
					// Send the message (will be done effectively only if outgoing status of the interface is active)
					tmpInterface.sendMSDDescription ( description );
					
					// Remove the last hop (list starts at 0) and continue
					description.getPropaInfo().getHops().remove( lastPos );
				}
			}
		//}
	}
	
	
	/**
	 * Propagates an msd request received on an interface onto the other interfaces.<br>
	 * This method needs to be synchronized as it manipulates the interface list.<br>
	 * 
	 * @param srcId the source network the msd description is coming from
	 * @param bridgeInterfaceId the id of the bridge interface
	 * @param request the msd request to propagate
	 */
	public void processIncomingMessage ( UbisecID srcId, int bridgeInterfaceId, MSDRequest request )
	{
		board.info ("- Received a request from interface on " + bridgeInterfaceId );
		//board.info("- Received a request from interface on " + bridgeInterfaceId );
		
		// When receiving a message, the MSD Bridge first checks that it has not yet forwarded the same
		// message. If so, the message is just discarted.
		if ( request.getPropaInfo()!=null ) {
			List hops = request.getPropaInfo().getHops();
			for (int x=0; x<hops.size();x++)
			{
				Hop tmpHop = (Hop)hops.get(x);
				if ( MSDUtil.getInstance().compareID(tmpHop.getBridgeID(), this.bridgeId) )
				{
					// already forwarded 
					board.info ("   - Already forwarded this request, I dump it" );
					return;
				}
			}
		}
		
		// First, check all networks this bridge is connected to, and propagate the message
		// (except the one the message is coming from).
		ThreadBridgingInterface tmpInterface;
		//synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++)
			{
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (tmpInterface.getBridgeInterfaceId() == bridgeInterfaceId)
				{					
					// this is the incoming interface, no forwarding
					board.info ("- Received from interface on " + tmpInterface.getIfname() + " " + tmpInterface.getGroupAddress() + ":" + tmpInterface.getGroupPort() );
					//board.info ("- Received from interface on " + tmpInterface.getIfname() + " " + tmpInterface.getGroupAddress() + ":" + tmpInterface.getGroupPort() + " with interface id " + tmpInterface.getBridgeInterfaceId() );
				}
				else
				{
					// Before propagating the message, a new "hop" is added to the propagation list
					// Change the destination network in each hope for each interface (see note below)
					board.info ("- Forward the request on interface on " + tmpInterface.getIfname() + " " + tmpInterface.getGroupAddress() + ":" + tmpInterface.getGroupPort() );
					//board.info ("- Forward the request on interface on " + tmpInterface.getIfname() + " " + tmpInterface.getGroupAddress() + ":" + tmpInterface.getGroupPort() );
					Hop myHop = new HopImpl();
					myHop.setBridgeID(bridgeId);
					myHop.setSrcNetID(srcId);
					myHop.setDestNetID(tmpInterface.getNetId());
					myHop.setSecurity("nosecurity");
					myHop.setTime(Calendar.getInstance());
					if ( request.getPropaInfo()==null ) {
						PropagationInformation propagation = new PropagationInformationImpl();
						request.setPropaInfo( propagation );
					}
					// WARNING: here the size of the list is recorded, because the hop added may then be removed
					// This is because the bridge may send the message to more than one network so the latest 
					// hop (destination network) may be different
					int lastPos = request.getPropaInfo().getHops().size();
					
					// Check that the request has not yet been on the network (from another bridge)
					List hops = request.getPropaInfo().getHops();
					if ( hops!=null ) {
						Iterator it = hops.iterator();
						while( it.hasNext() ) {
							Hop tmpHop = (Hop)it.next();
							
							if ( MSDUtil.getInstance().compareID( tmpHop.getDestNetID(), tmpInterface.getNetId() ) 
									|| MSDUtil.getInstance().compareID( tmpHop.getSrcNetID(), tmpInterface.getNetId() )  ) {
								board.info(" - Received a MSDRequest that has already been on this network..."); 
								return;   
							}
						}
					}
					
					// Add the hop and send the description
					request.getPropaInfo().getHops().add ( myHop );
					
					//
					// ORI- context management
					//
					if (daemon.isContextEnabled())
					{
						//board.debug( "START CONTEXT MANAGEMENT ... " );
						
						//Increment the #hop in the CtxInfo
						List list = request.getCtxInfo().getCINet().getCITable();
						int index = ContextUtil.getInstance().getIndex( NetworkTemplate.HOP, list );
						ContextElement elt = (ContextElement)list.get(index);
						BigInteger big = elt.getVal().getInt().add( BigInteger.valueOf(1) );
						elt.getVal().setInt( big );
						list.set(index, elt );
						
						// Context and rules for the network provided by the manager...
						InformationComponent destInfo = localMgr.getManager( myHop.getDestNetID().getName() );
						ContextInformationTable netCtx_dest = destInfo.getCtxInfo();
						//ContextRulesTree rulesCtx_dest = destInfo.getCtxRules();
						
						// Context and rules for the network provided by the request/user...
						// Check for the #hop 
						// 		ciUserNet and crUserNet
						ContextInformationTable ciUserNet = null;
						if ( request.getCtxInfo()!=null )
						    ciUserNet = request.getCtxInfo().getCINet();
						
						ContextRulesTree crUserNet = null;
						if ( request.getCtxRules()!=null ) 
						    crUserNet = request.getCtxRules().getCRNet();
						
						
						//
						// Network context and user request matching
						//
						
						ContextMatcher matcher = new ContextMatcher(request, ciUserNet, netCtx_dest, crUserNet );
						if (!matcher.matchRequestNetwork())
						{	
						    board.info("MSDBridge: the request is NOT forwarded onto the network");
						}
						else
						{
						    board.info("MSDBridge: the request is forwarded onto the network");
							
						    board.info(" SEND TO "+tmpInterface.getGroupAddress()+" "+tmpInterface.getGroupPort()+" "+tmpInterface.getIfname());
							board.debug("**************** tmpInterface.sendMSDRequest ( request ); *********");
							// Send the message (will be done effectively only if outgoing status of the interface is active)
							tmpInterface.sendMSDRequest ( request );
							board.info(" SEND TO "+tmpInterface.getGroupAddress()+" "+tmpInterface.getGroupPort()+" "+tmpInterface.getIfname());
							
							//board.debug(" SEND TO "+tmpInterface.getGroupAddress()+" "+tmpInterface.getGroupPort()+" "+tmpInterface.getIfname());
						}
						
						// Remove the last hop (list starts at 0) and continue in the loop
						request.getPropaInfo().getHops().remove( lastPos );
						
						// Decrement the #hop in the CtxInfo
						list = request.getCtxInfo().getCINet().getCITable();
						index = ContextUtil.getInstance().getIndex( NetworkTemplate.HOP, list );
						elt = (ContextElement)list.get(index);
						big = elt.getVal().getInt().subtract( BigInteger.valueOf(1) );
						elt.getVal().setInt( big );
						list.set(index, elt );
						
						//board.debug( "END CONTEXT MANAGEMENT ... " );
					}
					else {
					    // default case, always forward and remove the last hop 
						board.info(" SEND TO "+tmpInterface.getGroupAddress()+" "+tmpInterface.getGroupPort()+" "+tmpInterface.getIfname());
						tmpInterface.sendMSDRequest ( request );
						request.getPropaInfo().getHops().remove( lastPos );
					}
				}
			}
		//}
	}
	
	/**
	 * Propagates an MSDA message received on an interface onto the expected interface.<br>
	 * 
	 * @param srcId
	 * @param bridgeInterfaceId
	 * @param message
	 */
	 public void processIncomingMessage(  UbisecID srcId, int bridgeInterfaceId, MSDMessage message ) {
		
		//board.info("- RESPONSE REQUEST ! Received a message from interface on " + bridgeInterfaceId );
		// Verify the propagationInformation
		//board.info ("- RESPONSE REQUEST ! Received a message from interface on " + bridgeInterfaceId );
		
		if ( message.getBackwardPath()==null ) {
			board.info( "Received a MSDMessage without backwardpath..." );
			board.info( "Received a MSDMessage without backwardpath..." );
			return;
		}
		
		// Search if this MSDBridge has forwarded the corresponding MSDRequest
		// whose propagation information has been copied
		Hop tmpHop = null;
		List hops = message.getBackwardPath().getHops();
		for (int x=0; x<hops.size(); x++)
		{
			tmpHop = (Hop)hops.get(x);
			if ( MSDUtil.getInstance().compareID( tmpHop.getBridgeID(), this.bridgeId ) ) {
				// The hop to consider
				break;
			}
		}
		
		// Search among the active network interfaces
		// if one corresponds to the DestNetwork
		// and another corresponds to the SrcNetwork
		
		if ( tmpHop!=null ) {
			
			ThreadBridgingInterface srcInterface = null;
			ThreadBridgingInterface destInterface = null;
			//synchronized (activeInterfacesVector) {
				for (int x=0; x<activeInterfacesVector.size(); x++)
				{
					ThreadBridgingInterface tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
					if ( srcInterface==null && MSDUtil.getInstance().compareID( tmpInterface.getNetId(), tmpHop.getSrcNetID() ) ) {
						srcInterface = tmpInterface;
					}
					
					if ( destInterface==null && MSDUtil.getInstance().compareID( tmpInterface.getNetId(), tmpHop.getDestNetID() ) ) {
						destInterface = tmpInterface;
					}
				}
			//}
			// This Bridge has the corresponding network interfaces
			if ( srcInterface!=null && destInterface!=null ) {
				// Propagates the MSDMessage on the srcInterface because this message follows the backward path
				// To prevent another bridge from re-forward it, remove the hop
				message.getBackwardPath().getHops().remove( tmpHop );
				board.info(" RESPONSE REQUEST sent over "+destInterface.getNetId().getName() + " and req number " + message.getRespCode() );
				destInterface.sendMSDMessage( message );
				//board.info( message.getContent().getDesc().getDescID().getName() + " time="+System.currentTimeMillis() );
			} 
			
		} else {
			// Search TODO - verify only the network interface if 
			//               we can know that there is not another Bridge
			// 2 cases : there is more than 1 bridge with the sames network interfaces
			//           or the previous bridge has disappeared
			board.info(" This Bridge has not forwarded the MSD Request... ");
		}
		
	}
	
	
	/**
	 * Sets the bridging status for network interface
	 * 
	 * @param netId the network id of the network to setup
	 * @param activateIn true if bridge should accept incoming messages from this network, false otherwise
	 * @param activateOut true if bridge should propagate outgoing messages to this network, false otherwise	 
	 * @see org.msda.components.interfaces.MSDABridgeInterface#setBridgingStatus(org.msda.msddescription.UbisecID, boolean, boolean)
	 *
	 */
	public void setBridgingStatus ( UbisecID netId, boolean activateIn, boolean activateOut ) throws MSDException
	{
		ThreadBridgingInterface tmpInterface;
		synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++)
			{
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (tmpInterface.getNetId().equals(netId)) {
					tmpInterface.setOutgoingStatus(activateOut);
					tmpInterface.setIncomingStatus (activateIn);
					return;
				}
			}
		}
		Throwable e = new Throwable ();
		throw new MSDException ("MSDBrigde::setBridgingStatus : Interface not found: " + netId , e);
	}
	
	/**
	 * Returns the bridging status for incoming packets on the specified network
	 * 
	 * @param netId the network to check
	 * @return true if incoming messages accepted, false otherwise
	 * @see org.msda.components.interfaces.MSDABridgeInterface#getIncomingBridgingStatus(org.msda.msddescription.UbisecID)
	 *
	 */
	public boolean getIncomingBridgingStatus ( UbisecID netId ) throws MSDException
	{
		ThreadBridgingInterface tmpInterface;
		synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++)
			{
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (tmpInterface.getNetId().equals(netId)) {
					return tmpInterface.incomingStatus();
				}
			}
		}
		Throwable e = new Throwable ();
		throw new MSDException ("MSDBrigde::getIncomingBridgingStatus : Interface not found: " + netId , e);
	}
	
	/**
	 * Returns the bridging status for outgoing packets on the specified network
	 * 
	 * @param netId the network to check
	 * @return true if messages propagated to this network, false otherwise
	 * @see org.msda.components.interfaces.MSDABridgeInterface#getOutgoingBridgingStatus(org.msda.msddescription.UbisecID)
	 *
	 */
	public boolean getOutgoingBridgingStatus ( UbisecID netId ) throws MSDException
	{
		ThreadBridgingInterface tmpInterface;
		synchronized (activeInterfacesVector) {
			for (int x=0; x<activeInterfacesVector.size(); x++)
			{
				tmpInterface = (ThreadBridgingInterface)activeInterfacesVector.get(x);
				if (tmpInterface.getNetId().equals(netId)) {
					return tmpInterface.outgoingStatus();
				}
			}
		}
		Throwable e = new Throwable ();
		throw new MSDException ("MSDBrigde::getOutgoingBridgingStatus : Interface not found: " + netId , e);
	}
	
	
	/*
	 *  (non-Javadoc)
	 * @see org.msda.components.MSDBridgeInterface#updateInfoComponent(int, org.msda.msd.InformationComponent)
	 */
	public void updateInfoComponent(int typeComponent, InformationComponent info) {
		switch (typeComponent)
		{
		case InfoType.Component.BRIDGE:
			localMgr.addBridge(info);
			localMgr.addComponent(info);
		break;
		case InfoType.Component.DEVICE:
			localMgr.addComponent(info);
		break;
		case InfoType.Component.MANAGER:
			localMgr.addManager(info);
			localMgr.addComponent(info);
		break;
		}
	}	

	/*
	 *  (non-Javadoc)
	 * @see org.msda.components.MSDBridgeInterface#removeInfoComponent(int, org.msda.msd.InformationComponent)
	 */
	public void removeInfoComponent(int typeComponent, InformationComponent info) {
		switch (typeComponent)
		{
		case InfoType.Component.BRIDGE:
			localMgr.removeBridge(info);
		break;
		case InfoType.Component.DEVICE:
			localMgr.removeComponent(info);
		break;
		case InfoType.Component.MANAGER:
			localMgr.removeManager(info);
		break;
		}
	}	
	
	
	
	//
	// Needed for GUI
	//
	
	/**
	 * Returns the Vector of network interfaces
	 * 
	 * @return Returns the Vector of network interfaces (NetworkConfig, ...).
	 */
	public Vector getNetworkInterfacesVector() {
		return networkInterfacesVector;
	}
	
	
} // class MSDBridge
