/*
 * Created on May 17, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.bridge;

import org.msda.components.interfaces.*;
import org.msda.framework.FrameworkOutput;
import org.msda.msd.MSDDescription;
import org.msda.msd.MSDMessage;
import org.msda.msd.MSDRequest;
import org.msda.msd.NetworkConfig;
import org.msda.msd.UbisecID;
import org.msda.msd.impl.UbisecIDImpl;
import org.msda.util.*;



/**
 * This class binds together a specific group & interface, and manages 
 * communication (send/receive) to/from this "interface". <br>
 * When a message is received, it is forwarded to the 
 * bridge that decides if it should be forwarded and how to modify it.
 * 
 * @author raverdy, delachap - INRIA ARLES
 */
public class ThreadBridgingInterface extends Thread implements CommunicatorCallback {
    
    private int bridgeInterfaceId;
    private String ifname;
    private String groupAddress;
    private int groupPort;
    private UbisecID netId;
    private String IPAddress;
    
    private CommunicatorInterface communicator;
    
    private boolean activeOut;
    private boolean activeIn;
    
    private FrameworkOutput board;
    private MSDBridge bridge;
    
    /**
     * Creates a BridgingInterface that will connect the bridge with the msd manager and other bridges in a network.<br>
     * The thread is not started in the constructor.
     *
     * @param bridge
     * @param board 
     * @param networkConfig that contains <p>
	 *  		"netname" name of the network (for id purposes) <br>
	 * 			"ifname" name of the network interface to use (eth0...) <br>
	 * 			"groupAddress" multicast group address to use for communicating with the msd manager <br>
	 * 			"groupPort" port number to use for communicating with the msd manager 
	 * @param IPAddress
	 * @param interfaceId
     */		
    public ThreadBridgingInterface ( MSDBridge bridge, NetworkConfig networkConfig, String IPAddress, int interfaceId) throws MSDException
    {
        this.bridge = bridge;
		this.board = org.msda.framework.FrameworkMain.msdaOutput;
        
        this.ifname = networkConfig.getInterface();
        this.groupAddress = networkConfig.getDataGroup();
        this.groupPort = networkConfig.getDataPort().intValue();
        this.IPAddress = IPAddress;
        this.bridgeInterfaceId = interfaceId;
        this.activeIn = true;
        this.activeOut = true;
        //
        netId = new UbisecIDImpl();
        netId.setName( networkConfig.getName() );
        //
        // in case of problem, pass along their exception
        MSDANetworkInterface netif = bridge.getNetworkInterface();
        this.communicator = netif.registerCallback ( netId, ifname, groupAddress, groupPort, this );
    }
    
    
    public void finalize ()throws Throwable
    {
		try {
			bridge = null;
			groupAddress = null;
			IPAddress = null;
			netId = null;
			board = null;
			communicator.terminate();
		} catch (Exception e) {
		} finally {
			super.finalize ();
		}
    	
    }
    /**
     */
    synchronized public void terminate() {
    	try {
    		this.communicator.unregisterCallback(this);
    		communicator.terminate();
    	} 
    	catch (MSDException e)
		{
		}
    }
    
    /**
     * Returns the multicast group address
     * 
     * @return Returns the multicast group address.
     */
    public String getGroupAddress() {
        return groupAddress;
    }
    /**
     * Returns the multicast group port
     * 
     * @return Returns the multicast group port.
     */
    public int getGroupPort() {
        return groupPort;
    }
    /**
     * Returns the network interface name 
     * 
     * @return Returns the interface name.
     */
    public String getIfname() {
        return ifname;
    }
    /**
     * Returns the network UbisecID
     * 
     * @return Returns the network UbisecID.
     */
    public UbisecID getNetId() {
        return netId;
    }
    /**
     * Returns the bridge interface id.
     * 
     * @return Returns the bridgeInterfaceId.
     */
    public int getBridgeInterfaceId() {
        return bridgeInterfaceId;
    }
    /**
     * 
     * @see org.msda.components.interfaces.CommunicatorCallback#newDescription(org.msda.msddescription.MSDDescription)
     *
     */
    public void newDescription (MSDDescription description )
    {
        // Only process the message if the interface is set to process incoming messages
        if (activeIn) {
            bridge.processIncomingMessage ( this.netId, this.bridgeInterfaceId, description );
        }
        return;
    }
    
    /**
     * 
     * @see org.msda.components.interfaces.CommunicatorCallback#newRequest(org.msda.msdrequest.MSDRequest)
     *
     */
    public void newRequest (MSDRequest request )
    {
    	// Only process the message if the interface is set to process incoming messages
        if (activeIn) {
            bridge.processIncomingMessage ( this.netId, this.bridgeInterfaceId, request );
        }
        return;
    }
    
    /**
     * @see org.msda.components.interfaces.CommunicatorCallback#responseRequest(org.msda.msd.MSDMessage)
     */
    public void responseRequest ( MSDMessage message ) {
        
        // Only process the message if the interface is set to process incoming messages
        if (activeIn) {
            bridge.processIncomingMessage(this.netId, this.bridgeInterfaceId, message);
        }
        return;
    }
    
    /**
	 * Receives an incoming msd message of type InformationComponent.<br>
	 * The msd message contains the IP address and the port to process a remote access.
	 * 
	 * @param message received from the msd group
	 * 
	 * @see org.msda.components.interfaces.CommunicatorCallback#getInformationComponents(org.msda.msd.MSDMessage)
	 */
     synchronized public void controlMessage(MSDMessage message) {
//        
//        board.debug("MSDBridge::controlMessage : Processing of information about another component of the multicast group." );
//        
//        switch( message.getContent().getControl().getInfoComponent().getComponentType().intValue() ) {
//        
//        case InfoType.Component.BRIDGE_COMPONENT:
//        case InfoType.Component.BRIDGE_MANAGER_COMPONENT:  
//            
//            // search if this is the returned message send by this bridge
//            InformationComponent info = message.getContent().getControl().getInfoComponent();
//            if ( MSDUtil.getInstance().compareID( info.getComponentID(), bridge.getBridgeID() ) ) {
//                // do nothing, because this message was sent by this bridge
//                return;
//            }
//        }
//        
//        int res = bridge.getLocalMgr().addComponent( message );
//        MSDMessage myMessage = null;
//        switch( res ) {
//        
//        case 0: break;
//        case 2: break;
//        case 1:
//        case 3:
//        case 4:
//        case 5:
//            
//            myMessage = bridge.getLocalMgr().createMSDMessage( InfoType.Component.BRIDGE_COMPONENT, bridge.getBridgeID(), bridge.getIpAddress(), bridge.getRemoteAccessPort() );
//        
//            // ADLC TMP - How to know this device is bridge and manager ?
//            //myMessage = bridge.getLocalMgr().createMSDMessage( InfoType.Component.BRIDGE_MANAGER_COMPONENT, bridge.getBridgeID(), bridge.getIpAddress(), bridge.getRemoteAccessPort() );
//            
//            // add all the active network interfaces
//            List list = myMessage.getContent().getControl().getInfoComponent().getNetworks();
//            for( int i=0; i<bridge.getActiveInterfacesVector().size(); i++) { 
//                
//                ThreadBridgingInterface interf = (ThreadBridgingInterface)bridge.getActiveInterfacesVector().get(i);
//                list.add( interf.getNetId() );
//            }
//            
//            // complete the InformationComponent
//            boolean found = false;
//            for(int i=0;i<list.size();i++) {
//            	UbisecID id = (UbisecID)list.get(i);
//            	found = false;
//	            for(int j=0; j<bridge.getNetworkInterfacesVector().size(); j++) {
//	            	
//	            	NetworkConfig netConf = (NetworkConfig) bridge.getNetworkInterfacesVector().get(j);
//	            	// TODO replace with compareID
//	            	if ( netConf.getName().equals( id.getName() ) ) {
//		            	//add all the IPAddress and ports..
////		                myMessage.getContent().getInfoComponent().getRemoteAccessPorts().add( netConf.getRemoteAccessPort() );
////		                myMessage.getContent().getInfoComponent().getIPAddresses().add( netConf.getIPAddress() );
//		                found = true;
//		                break;
//	            	}
//	            }
//	            if ( !found ) {
//	            	board.debug( "MSDBridge::controlMessage: problem with Network interfacees");
//	            	myMessage.getContent().getControl().getInfoComponent().getRemoteAccessPorts().add( BigInteger.valueOf( 8080 ) );
//	            	myMessage.getContent().getControl().getInfoComponent().getIPAddresses().add( "127.0.0.1" );
//	            }
//            }
//            
//            
//            
//            board.debug( "MSDBridge::controlMessage: send a message to the multicast group..." );
//            
//            sendMSDMessage( myMessage );
//        }
    }
    
    /**
     * Sends an msd description on the interface/network associated with the BridgingInterface
     * if and only if the interface is active (outgoing)
     * 
     * @param description the meta description to send on the interface/network
     * @return true if description was sent correctly, false otherwise
     */
    public boolean sendMSDDescription (MSDDescription description )
    {
        if (!activeOut)
        {
            return false;
        }
        else 
            if (communicator == null)
            {
                // here the BridgingInterface was not correctly initialized and should be removed
                board.warn("cannot send msd description, communicator not initialized" );
                return false;
            }
            else
            {
                try 
                {
                    this.communicator.sendMSDDescription ( description );
                }
                catch (MSDException e)
                {
                    // Problem when sending the description.
                    board.debug ("error when sending a msd descripion" );
                    return false;
                }
            }
        return true;
    }
    
    /**
     * Sends an msd request on the interface/network associated with the BridgingInterface
     * if and only if the interface is active (outgoing)
     * 
     * @param request the msd request to send on the interface/network
     * @return true if request was sent correctly, false otherwise
     */
    public boolean sendMSDRequest (MSDRequest request )
    {
        if (!activeOut)
        {
            board.warn ("cannot send request, outgoing interface inactive" );
            return false;
        }
        else 
            if (communicator == null)
            {
                // here the BridgingInterface was not correctly initialized and should be removed
                board.warn ("cannot send request, communicator not initialized" );
                return false;
            }
            else
            {
                try 
                {
                    this.communicator.sendMSDRequest ( request );
                }
                catch (MSDException e)
                {
                    // problem when sending the request.
                    board.debug ("error when sending a request" );
                    return false;
                }
            }
        return true;
    }
    
    /**
     * Sends an msd message on the interface/network associated with the BridgingInterface
     * if and only if the interface is active (outgoing)
     * 
     * @param message the msd message to send on the interface/network
     * @return true if message was sent correctly, false otherwise
     */
    public boolean sendMSDMessage( MSDMessage message ) {
        
        if (!activeOut)
        {
            board.warn ("cannot send message, outgoing interface inactive" );
            return false;
        }
        else 
            if (communicator == null)
            {
                // here the BridgingInterface was not correctly initialized and should be removed
                board.warn ("cannot send message, communicator not initialized" );
                return false;
            }
            else
            {
                try 
                {
                    this.communicator.sendMSDMessage( message );
                }
                catch (MSDException e)
                {
                    // problem when sending the request.
                    board.debug ("error when sending a message" );
                    return false;
                }
            }
        return true;
    }
    
    /**
     * Returns the propagation status of the interface for incoming messages.
     * 
     * @return status of the interface. true if the interface should process incoming messages, false otherwise.
     */		
    synchronized public boolean incomingStatus ()
    {
        return activeIn;
    }
    
    /**
     * Returns the propagation status of the interface for outgoing messages.
     * 
     * @return status of the interface. true if the interface should forwards messages on this interface, false otherwise.
     */		
    synchronized public boolean outgoingStatus ()
    {
        return activeOut;
    }
    
    
    /**
     * Sets the propagation status of the interface for incoming messages.
     * 
     * @param val status of the interface. true if the interface should process incoming messages, false otherwise.
     */
    public void setIncomingStatus (boolean val)
    {
        activeIn = val;
        return;
    }
    
    
    /**
     * Set the propagation status of the interface for outgoing messages.
     * 
     * @param val status of the interface. true if the interface should forwards messages on this interface, false otherwise.
     */
    public void setOutgoingStatus (boolean val)
    {
        activeOut = val;
        return;
    }		
}
