/*
 * Created on Aug 26, 2004
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.core;


import org.msda.components.interfaces.CommunicatorCallback;
import org.msda.components.interfaces.CommunicatorInterface;
import org.msda.components.interfaces.MSDANetworkInterface;
import org.msda.msd.*;
import org.msda.msd.impl.MSDMessageImpl;
import org.msda.msd.impl.MessageContentImpl;
import org.msda.util.*;
import org.msda.util.comm.MulticastGroup;
import org.msda.util.comm.MulticastGroupCallback;


import java.math.BigInteger;
import java.util.*;

/**
 * This class implements the MSD network interface that manages low level network management features. 
 * 
 * @author raverdy
 *
 */
public class MSDNetwork implements MSDANetworkInterface {
	
	private Vector communicationManagerList;	
	private int globalComponentCounter = 1;
	private UbisecID networkID;
	private MSDDaemon daemon;
	
	/**
	 * Creates the MSD network component
	 * @param the log
	 */
	public MSDNetwork(MSDDaemon daemon)
	{
		this.daemon = daemon;
		this.networkID = daemon.getUbisecID();
		communicationManagerList = new Vector();
	}
	
	/**
	 * @see java.lang.Object#finalize()
	 */
	public void finalize () throws Throwable
	{
		CommunicatorInterface commMgr;
		try {
			while (communicationManagerList.size()>0) {
				commMgr = (CommunicatorInterface)communicationManagerList.remove(communicationManagerList.size()-1);
				commMgr.terminate();
				commMgr = null;
			}
		} catch (Exception e) {
		} finally {
			communicationManagerList = null;
			super.finalize ();
		}
	}
	
	
	/**
	 * Creates an entry for communicating with an MSD group (MSD manager and bridges) and returns the interface to this MSD group.
	 * With the usual parameters (ip address and port) the caller should provide a callback interface.
	 * 
	 * @param netId unique ID of the MSD component (MSND Manager or MSD Bridge) that is added to messages sent to the network
	 * @param ifname network interface name to use for connecting to the network
	 * @param groupAddress multicast group address to use for connecting to the network
	 * @param groupPort port to use for connecting to the network
	 * @param callback a CommunicatorCallback interface used for forwarding messages received from the network
	 * @return the proxy interface for sending meta descriptions/requests on the network
	 * @throws MSDException
	 * 
	 * @see org.msda.components.interfaces.MSDANetworkInterface#registerCallback(org.msda.msddescription.UbisecID, java.lang.String, java.lang.String, int, org.msda.components.CommunicatorCallback)
	 */
	public CommunicatorInterface registerCallback (UbisecID netId, String ifname, String groupAddress, int groupPort, CommunicatorCallback callback) throws MSDException
	{
		CommunicatorInterface commMgr;
		// first check with the existing msd group if one was already created with these parameters.
		// if yes, register the new client with this msd group
		// if no, create the msd group and register the client
		
		synchronized (communicationManagerList) {
			for (int x=0; x<communicationManagerList.size(); x++)
			{
				commMgr = (CommunicatorInterface)communicationManagerList.get(x);
				if (commMgr.networkInterface().equals(ifname) && commMgr.address().equals(groupAddress) && (commMgr.port()==groupPort))
				{
					try {
						// register to an existing group
						commMgr.registerCallback (callback);
						return commMgr;
					}
					catch (MSDException e)
					{
						// TODO - what to do in case of an error ?
						// Could not register : commMgr.networkInterface() commMgr.address() commMgr.port()
					}
				}
			}
		}
		// create the group and register to it
		commMgr = startCommunicationManager ( netId, ifname, groupAddress, groupPort);
		commMgr.registerCallback (callback);
		return commMgr;
	}
	
	
	/**
	 * Starts a communication manager for a given msd group.
	 * The interface name, and group/port to use are provided to the msd network component that creates 
	 * the relevant <code>communicationManager</code>.
	 * 
	 * @param netId the network id associated with this msd group
	 * @param ifname the network interface name
	 * @param groupAddress the multicast group address to use
	 * @param groupPort the multicast group port to use
	 * @return returns the interface to the msd group (<code>CommunicatorInterface</code>)
	 * @throws MSDException throws an exception if the msd group could not be created.
	 */
	public CommunicatorInterface startCommunicationManager ( UbisecID netId, String ifname, String groupAddress, int groupPort ) throws MSDException
	{
		CommunicatorInterface commMgr;
		commMgr = new CommunicationManager ( globalComponentCounter++, netId, ifname, groupAddress, groupPort );
		communicationManagerList.add( commMgr );
		return commMgr;
	}
	
	public boolean stopCommunicationManager ( UbisecID netId )
	{
		return false;
	}
	
	/**
	 * This inner class implements the MSD group.
	 * Each MSD group has a communicator so a bridge will have different communicators, one for each
	 * "network interface" (interface, group, port) in order to send MSD messages (description or requests)
	 * In theory, a "MSD group" is created each physical LAN (must support multicast) but in practice, multiple 
	 * "MSD groups" can be created over the same physical LAN in order to do experiments.
	 * 
	 * @author raverdy
	 */
	
	public class CommunicationManager implements CommunicatorInterface, MulticastGroupCallback {
		UbisecID componentID;
		UbisecID netId;
		String ifname;
		String groupAddress;
		int groupPort;
		MulticastGroup mgroup;
		Vector communicationCallbackList;	
		
		/**
		 * Creates a communication manager according to the provided parameters.
		 * 
		 * @param componentCounter
		 * @param netId the network id associated with this msd group
		 * @param ifname the network interface name
		 * @param groupAddress the multicast group address to use
		 * @param groupPort the multicast group port to use
		 * @throws MSDException throws an exception if the msd group could not be created.
		 */
		public CommunicationManager (int componentCounter, UbisecID netId, String ifname, String groupAddress, int groupPort ) throws MSDException
		{
			this.netId = netId;
			// first I initialize the member variable
			communicationCallbackList = new Vector();
			this.componentID = daemon.getUbisecID();
			this.ifname = ifname;
			this.groupAddress = groupAddress;
			this.groupPort = groupPort;
			// now create the multicast group on the specified interface.
			// here do not handle the MSDException but pass it along to the caller
			mgroup = new MulticastGroup( this, ifname, groupAddress, groupPort);
		}
		
		/**
		 * 
		 */
		synchronized public void terminate ()
		{
			if (mgroup!=null) 
			{
				mgroup.terminate();
				mgroup = null;
			}
			communicationCallbackList = null;
		}
		
		/**
		 * Returns the number of callbacks registerred for this multicast group
		 * 
		 * @return the number of calbacks registered
		 * @see org.msda.components.interfaces.CommunicatorInterface#getCallbacksNumber()
		 */
		public int getCallbacksNumber()
		{
			return communicationCallbackList.size();
		}
		
		/**
		 * Returns the UbisecID for this network
		 * 
		 * @return the network UbisecID
		 * @see org.msda.components.interfaces.CommunicatorInterface#getNetworkID()
		 */
		public UbisecID	getNetworkID ()
		{
			return netId;
		}
		
		/**
		 * Returns the name of the network interface used by this network
		 * 
		 * @return the name of the network interface
		 * @see org.msda.components.interfaces.CommunicatorInterface#networkInterface()
		 */
		public String networkInterface()
		{
			return ifname;
		}
		
		/**
		 * Returns the multicast group address used by this network
		 * 
		 * @return the multicast group address
		 * @see org.msda.components.interfaces.CommunicatorInterface#address()
		 */
		public String address()
		{
			return groupAddress;
		}
		
		/**
		 * Returns the port number used by this network
		 * 
		 * @return the port number
		 * @see org.msda.components.interfaces.CommunicatorInterface#port()
		 */
		public int port()
		{
			return groupPort;
		}
		
		
		//
		// 			MulticastGroupCallback 
		//
		/**
		 * The MSD message holds a response for a discoveryRequest
		 * 
		 * @param message The MSD message received on the multicast group
		 */
		public synchronized void receivedMessage ( MSDMessage message )
		{
			CommunicatorCallback callback;
			for (int x=0; x<communicationCallbackList.size(); x++)
			{
				callback = (CommunicatorCallback)communicationCallbackList.get(x);
				switch (message.getType().intValue())
				{
				case InfoType.Message.BRIDGE_DESACTIVATION:
				case InfoType.Message.BRIDGE_ACTIVATION:
				case InfoType.Message.BEACON_DEVICE:
				case InfoType.Message.INFO_COMPONENT:
				case InfoType.Message.GET_INFO_COMPONENT:
				case InfoType.Message.EXIST_MANAGER:
				case InfoType.Message.BEACON_MANAGER:
				case InfoType.Message.BEACON_BRIDGE:
					callback.controlMessage(message);
				break;
				case InfoType.Message.SERVICE_UPDATE:
					callback.newDescription ( message.getContent().getDesc() );
				break;
				case InfoType.Message.REQUEST_RESPONSE:
					callback.responseRequest(message);
				break;
				case InfoType.Message.SERVICE_REQUEST:
					callback.newRequest(message.getContent().getRequest());
				break;
				case InfoType.Message.REMOTE_ACCESS:
				case InfoType.Message.REMOTE_ACCESS_OK:
					//case InfoType.Message.REMOTE_ACCESS_ERROR:
					callback.responseRequest(message);
				//					maybe not that...
				//				    
				break;
				case InfoType.Message.INVALID:
				default:
					// Invalid/Unknow message type : message.getType().intValue() 
					x=communicationCallbackList.size();
				break;
				}
			}
		}
		
		//
		// 			CommunicatorInterface 
		//
		
		/**
		 * Registers a callback interface to receive messages from the network
		 * 
		 * @param callback the callback interface to register
		 * @throws MSDException throws an exception if the interface could not be registered
		 * @see org.msda.components.interfaces.CommunicatorInterface#registerCallback(org.msda.components.CommunicatorCallback)
		 */
		public void registerCallback(CommunicatorCallback callback)
		{
			communicationCallbackList.add(callback);
		}
		
		/**
		 * Unregisters a callback interface
		 * 
		 * @param callback the callback interface to unregister
		 * @throws MSDException throws an exception if the interface could not be registered
		 * @see org.msda.components.interfaces.CommunicatorInterface#registerCallback(org.msda.components.CommunicatorCallback)
		 */
		synchronized public void unregisterCallback(CommunicatorCallback callback)
		{
			CommunicatorCallback tmpCallback;
			for (int x=0; x<communicationCallbackList.size(); x++)
			{
				tmpCallback = (CommunicatorCallback)communicationCallbackList.get(x); 
				if (tmpCallback == callback)
				{
					communicationCallbackList.removeElementAt(x);
					x--;
				}
			}
			// if nobody is listenning anymore, terminate this multicast group...
			if (communicationCallbackList.size()==0)
			{
				synchronized (communicationManagerList) {
					for (int x=0; x<communicationManagerList.size(); x++) 
					{
						CommunicationManager tmpCommMgr;
						tmpCommMgr = (CommunicationManager)communicationManagerList.get(x);
						if (tmpCommMgr.communicationCallbackList.size()==0)
						{
							tmpCommMgr.terminate();
							communicationManagerList.removeElementAt(x);
							x--;
						}
					}
				}
			}
		}
		
		/**
		 * Sends an MSD description onto the network
		 * 
		 * @param description the msd description to send
		 * @throws MSDException throws an exception if the message could not be sent onto the network
		 * @see org.msda.components.interfaces.CommunicatorInterface#sendMSDDescription(org.msda.msd.MSDDescription)
		 */
		public void sendMSDDescription(MSDDescription description) throws MSDException {
			MSDMessage newMess = new MSDMessageImpl();
			MessageContent content = new MessageContentImpl();
			newMess.setContent(content);
			content.setDesc(description);
			newMess.setBackwardPath(null);
			newMess.setChecksum(null);
			newMess.setRespCode(null);
			newMess.setType(BigInteger.valueOf(InfoType.Message.SERVICE_UPDATE));
			this.sendMSDMessage(newMess);
		}
		
		/**
		 * Sends an MSD request onto the network
		 * 
		 * @param request the msd request to send
		 * @throws MSDException throws an exception if the message could not be sent onto the network
		 * @see org.msda.components.interfaces.CommunicatorInterface#sendMSDRequest(org.msda.msd.MSDRequest)
		 */
		public void sendMSDRequest(MSDRequest request) throws MSDException {
			MSDMessage newMess = new MSDMessageImpl();
			MessageContent content = new MessageContentImpl();
			newMess.setContent(content);
			content.setRequest(request);
			newMess.setBackwardPath(null);
			newMess.setChecksum(null);
			newMess.setRespCode(null);
			newMess.setType(BigInteger.valueOf(InfoType.Message.SERVICE_REQUEST));
			this.sendMSDMessage(newMess);
		}
		
		/**
		 * Sends an MSD control message onto the network
		 * 
		 * @param message
		 * @throws MSDException
		 * @see org.msda.components.interfaces.CommunicatorInterface#sendControlMessage(org.msda.msd.MSDMessage)
		 */
		public void sendControlMessage(MSDMessage message) throws MSDException {
			this.sendMSDMessage(message);
		}
		
		
		/**
		 * Sends an MSD message (response) onto the network
		 * 
		 * @param message
		 * @throws MSDException
		 * @see org.msda.components.interfaces.CommunicatorInterface#sendResponse(org.msda.msd.MSDMessage)
		 */
		public void sendResponse(MSDMessage message) throws MSDException {
			this.sendMSDMessage(message);
		}
		
		/**
		 * Sends an MSD message onto the network
		 * 
		 * @param message
		 * @throws MSDException
		 * @see org.msda.components.interfaces.CommunicatorInterface#sendMSDMessage(org.msda.msd.MSDMessage)
		 */
		public void sendMSDMessage(MSDMessage message) throws MSDException {
			message.setSender(this.componentID);
			mgroup.sendMSDMessage(message);
			if (InfoType.STACK_PER_DEVICE)
			{
				// TODO should send back the method to all local clients except the sender
			}
		}
		
	}
}

