/*
 * Created on Aug 27, 2004
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.interfaces;

import org.msda.msd.*;
import org.msda.util.*;

/**
 * This interface is used to access the network and send msd messages (descriptions, requests, ...)
 * on the network. It is also used to get information about the networking parameters.<p>
 * This interface is implemented by the inner class <code>CommunicationManager</code> 
 * in <code>MSDNetwork</code>. 
 * 
 * @author raverdy
 */
public interface CommunicatorInterface {

	/**
	 * Returns the name of the network interface used by this network
	 * 
	 * @return the name of the network interface
	 */
	public String networkInterface();
	
	/**
	 * Returns the multicast group address used by this network
	 * 
	 * @return the multicast group address
	 */
	public String address();
	
	/**
	 * Returns the port number used by this network
	 * 
	 * @return the port number
	 */
	public int port();

	/**
	 * Returns the number of callbacks registerred for this multicast group
	 * 
	 * @return the number of calbacks registered
	 */
	public int getCallbacksNumber();

	/**
	 * Returns the UbisecID for this network
	 * 
	 * @return the network UbisecID
	 */
	public UbisecID	getNetworkID ();

	/**
	 * 
	 *
	 */
	public void terminate ();

	/**
	 * Sends an MSDA description onto the network
	 * 
	 * @param description the msd description to send
	 * @throws MSDException throws an exception if the message could not be sent onto the network
	 */	
	public void sendMSDDescription ( MSDDescription description ) throws MSDException;

	/**
	 * Sends an MSDA request onto the network
	 * 
	 * @param request the MSDA request to send
	 * @throws MSDException throws an exception if the message could not be sent onto the network
	 */	
	public void sendMSDRequest ( MSDRequest request ) throws MSDException;

	/**
	 * Sends an MSDA message onto the network
	 * 
	 * @param message the MSDA message to send
	 * @throws MSDException throws an exception if the message could not be sent onto the network
	 */
	public void sendMSDMessage (MSDMessage message) throws MSDException;
	
	/**
	 * Sends a control message onto the network
	 * 
	 * @param message the MSDA message to send
	 * @throws MSDException throws an exception if the message could not be sent onto the network
	 */
	public void sendControlMessage (MSDMessage message) throws MSDException;

	/**
	 * Sends an MSDA message containing a response to a discovery request onto the network
	 * 
	 * @param message the MSDA message to send
	 * @throws MSDException throws an exception if the message could not be sent onto the network
	 */
	public void sendResponse (MSDMessage message) throws MSDException;

	/**
	 * Registers a callback interface to receive messages from the network
	 * 
	 * @param callback the callback interface to register
	 * @throws MSDException throws an exception if the interface could not be registered
	 */
	public void registerCallback(CommunicatorCallback callback) throws MSDException;

	/**
	 * Unregisters a callback interface
	 * 
	 * @param callback the callback interface to unregister
	 * @throws MSDException throws an exception if the interface could not be registered
	 */
	public void unregisterCallback(CommunicatorCallback callback) throws MSDException;
}
