/*
 * Created on Aug 20, 2004
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.manager;


import java.util.*;
import java.math.*;

import org.msda.components.core.MSDDaemon;
import org.msda.components.interfaces.*;
import org.msda.components.manager.client.ClientConnectionMgr;
import org.msda.components.manager.object.*;
import org.msda.components.utils.LocalComponentsMgr;
import org.msda.context.*;
import org.msda.framework.FrameworkOutput;
import org.msda.gui.GUINotification;
import org.msda.msd.*;
import org.msda.msd.impl.*;
import org.msda.util.*;
import org.msda.util.comm.NetworkUtils;
import org.msda.util.security.SecurityUtil;

import secureMulticast.server.*;


/**
 * This class implements the MSD manager in charge of <br>
 *     (1) interacting with specific SD plugins to collect service announcements and perform discovery requests and service access,<br> 
 * and (2) interact with MSD bridges to propagate the requests. <br>
 * This is the central component of the MSD platform.
 * <p>
 * 
 * @author raverdy, delachap
 */
public class MSDManager implements MSDAManagerInterface, PluginCallback, CommunicatorCallback, RequestInterface {
    
	// interfaces to the core and network components of the msd middleware
	private MSDDaemon daemon;
	private MSDANetworkInterface networkInterface;
	private CommunicatorInterface communicator;
	// pool of threads manager
	private MSDManagerPool managerPool;
	// local description manager (local and remote vectors)
	private ThreadDescriptionMgr descManager;
	// context manager
	private MSDContextManager ctxManager;
	// cache manager for the informations (UbisecID, IP, port) of local components
	private LocalComponentsMgr localMgr;
	// pending requests manager
	private PendingRequestMgr pendingRequestMgr;
	
	// informations of the manager
	private String name;
	private String netname;
	private String ifname;
	private UbisecID managerId;
	private String IPAddress;
	private int remoteAccessPort;
	private String groupAddress;
	private int groupPort;

	// local counters for the client, request, remote ID assigned by this manager	
	private int localCounterRemote = 0;
	private int localCounterRequest = 0;
	private int localCounterClient = 0;
	
	// Default value of the timeout defined in the configuration file
	private static long TimeoutGetServices = 2000;
	private static long TimeoutServiceAccess = 2000;
	
	public static final String FIELD_CLIENT_ID = "clientID";
	
	// trace in the same log
	private FrameworkOutput board;
	private Vector pluginList;
	private UbisecID netId;

	// cache for service access
	private SvcAccessCacheMgr svcAccCacheMgr;
	
	// socket clients connection manager
	private ClientConnectionMgr clientConnectionMgr;
	
	private boolean isActive;

	private boolean isContextEnabled;
	

	/**
	 * Initializes the msd manager based on the network parameters provided.
	 * 
	 * @param netName 		name of the network to use for msd
	 * @param ifname 		interface to use for sending/receiving msd descriptions with MSD Bridges
	 * @param groupAddress 	multicast group address to use
	 * @param groupPort 	multicast group port to use
	 * @param clientPort 	port to use for accepting incoming client connections
	 * @param coreInterface	interface to the core component
	 * @param logger 		the log
	 * 
	 * @throws MSDException throws an exception if network unavailable
	 */
	
	public MSDManager( MSDDaemon daemon, ManagerConfig configMgr, NetworkConfig configNet, boolean isContextEnabled ) throws MSDException
	{		
		String networkName;
		
	    // Initializing from the startup
	    // Initializing the different member variables
		this.board = org.msda.framework.FrameworkMain.msdaOutput;
	    this.pluginList = new Vector();
	    this.pendingRequestMgr = new PendingRequestMgr();
	    this.svcAccCacheMgr = new SvcAccessCacheMgr();
	    
	    this.managerId = new UbisecIDImpl();
	    this.managerId.setName(configMgr.getName());
	    this.ifname = configNet.getInterface();
	    groupAddress = configNet.getDataGroup();
	    groupPort = configNet.getDataPort().intValue();

		board.info("MSDManager:: Creating an MSD manager for network " + configMgr.getNetworkName());

	    networkName = configMgr.getNetworkName();

	    // Instantiation of the context manager and creation of the network context
	    this.isContextEnabled = isContextEnabled;
	    if ( isContextEnabled ) {
        	this.ctxManager = new MSDContextManager();
        	String filenameCtx = configMgr.getNetworkCtx();
    	    ctxManager.setNetworkContextFile(filenameCtx);		   
        }
	    
	    // Initializing the timeouts with "type" in the configuration file...
	    String value = ParamsUtil.getInstance().getValue( InfoType.Config.GET_SERVICES_TIMEOUT, configMgr.getParams() );
	    if ( value!=null )	TimeoutGetServices = Long.parseLong( value );
	    value = ParamsUtil.getInstance().getValue( InfoType.Config.SERVICE_ACCESS_TIMEOUT, configMgr.getParams() );
	    if ( value!=null )	TimeoutServiceAccess = Long.parseLong( value );
	    
	    // get the interface to the network component
	    // and register with the communicator to send/receive msd descriptions.
	    // (note: the communicator will be created if it does not exist yet)
	    this.daemon = daemon;
	    networkInterface = daemon.getNetworkInterface();
	    netId = new UbisecIDImpl();
	    netId.setName( configNet.getName() );
	    localMgr = new LocalComponentsMgr();
	    
	    isActive = false;
	    
	    this.IPAddress = NetworkUtils.getIPAddress( ifname );
	    this.name = configMgr.getName();
	    this.netname = configMgr.getNetworkName();
	    this.remoteAccessPort = configMgr.getRemoteAccessPort().intValue();
	    
	    if ( this.IPAddress==null ) {
	        board.info("Config file error...no corresponding network interface...");
	        System.exit(1);
	        return;
	    }
	    this.descManager = new ThreadDescriptionMgr();
	    
	    if ( configMgr.getDirectMSDPort()!=null )
	        this.clientConnectionMgr = new ClientConnectionMgr( this, configMgr.getDirectMSDPort().intValue() );
	    else 
	        board.info("WARNING :: No direct access (socket) possible with this manager");
	    
	    board.info("* MSD Manager initialized on " + ifname + " with group " + groupAddress + ":" + groupPort );
	    
	    boolean activeSecurity = false;
	    if ( activeSecurity ) {
		    try {
		    	List securityParams = configNet.getSecurity().getParams();
		    	ParamsUtil paramUtil = new ParamsUtil();
		    	
				LKHserver s= new LKHserver(
						paramUtil.getValue( "LKHAlgorithm", securityParams ),
						0,
						paramUtil.getValue( "EncryptionAlgorithm", securityParams ),
						groupAddress,
						Integer.parseInt(paramUtil.getValue( "LKHServerPort", securityParams )),
						Integer.parseInt(paramUtil.getValue( "LKHServerPort", securityParams ))
						);
				SecurityUtil secUtil=SecurityUtil.getInstance(ifname);
				s.addLKHListener(secUtil);
				secUtil.setGroupKey(s.start());
		    } catch (Exception e) {
				throw new MSDException("The LKHserver couldn't start."+e, e);
			}
	    }
	}
	
	/* MSDManagerInterface methods */
	
	/**
	 * Activates the MSD manager
	 * 
	 * @return true if the MSD manager is active without problem, false otherwise
	 * 
	 * @see org.msda.components.interfaces.MSDAManagerInterface#startManager()
	 */
	public boolean startManager()
	{
		board.debug ("start the MSD Manager" );

		if (isActive)
			return true;
		
		// register with the multicast group to send/receive msd messages
		try {
			// create the multicast communicator
			this.communicator = networkInterface.registerCallback ( netId, ifname, groupAddress, groupPort, this );
		} catch (MSDException e) {
		    board.debug("* could not register with multicast group " + ifname + ":" + groupAddress + ":" + groupPort );
		    board.notifyException(e);
			return false;
		}
		
		// create the pool of thread
		this.managerPool = new MSDManagerPool( this, communicator );

		// activate all the SD plugins of this manager
    	PluginInterface tmpInterface;
    	for (int x=0; x<pluginList.size();x++)
    	{
    		tmpInterface = (PluginInterface)pluginList.get(x);
			tmpInterface.activate();
			this.managerPool.createNewTaskPlugin( tmpInterface ); 
    	}

		isActive = true;
    	return true;
	}

	/**
	 * Desactivates the MSD manager (e.g., when the daemon/device detects another
	 * MSD Manager in the network).
	 * 
	 * @see org.msda.components.interfaces.MSDAManagerInterface#stopManager()
	 */
	public boolean stopManager()
	{
	    board.debug ("* stop the MSD Manager" );

    	if (!isActive)
			return true;
		isActive = false;
		
		// desactivate all the SD plugins
		PluginInterface tmpInterface;
    	for (int x=0; x<pluginList.size();x++)
    	{
    		tmpInterface = (PluginInterface)pluginList.get(x);
    		// XXX PGR to create
    		// tmpInterface.desactivate();
    	}
    	// then unregister from the multicast group (the socket may close)
    	try {
			this.communicator.unregisterCallback(this);
		}
		catch (MSDException e) {
		    board.debug("Could not unregister from the callback");
		    board.notifyException(e);
		}
		return true;
	}
	
	/**
	 * Terminates the different threads attached to the msd manager and free resources.
	 * This method especially terminates 
	 * (1) the description manager that stores all the msd description
	 * and has a thread that reads/writes on the multicast group, 
	 * This method also terminates the different SD plugins and the client listeners already started. 
	 *
	 * TODO - completes  it !
	 */
	public void terminate ()
	{
	    board.debug("ERROR: MSDManager:terminate needs to be completed" );
		
		//
		if (descManager!=null)
			descManager.terminate();

	}

	/**
	 * Returns the unique ID of the MSD manager (UbisecID)
	 * 
	 * @return the MSD manager UbisecID
	 * @see org.msda.components.interfaces.MSDAManagerInterface#getMSDManagerID()
	 */
	public UbisecID	getMSDManagerID ()
	{
		return managerId;
	}
	
	/**
	 * Sets the unique ID of the MSD manager (UbisecID)
	 * 
	 * @param id the UbisecID of the MSD manager
	 * @see org.msda.components.interfaces.MSDAManagerInterface#setMSDManagerID(org.msda.msddescription.UbisecID)
	 */
	public void	setMSDManagerID (UbisecID id)
	{
	    managerId = id;
	}
	
	/**
	 * Returns the manager name 
	 * 
	 * @return	Returns the manager name 
     * @see org.msda.components.interfaces.MSDAManagerInterface#getName()
     */
    public String getName() {
        return name;
    }
    
    /**
	 * Returns the network name of the manager
	 * 
	 * @return	Returns the network name of the manager
     * @see org.msda.components.interfaces.MSDAManagerInterface#getNetworkName()
     */
    public String getNetworkName() {
        return netname;
    }
    
    /**
	 * Returns the remote access port of the manager
	 * 
	 * @return	Returns the remote access port of the manager
     * @see org.msda.components.interfaces.MSDAManagerInterface#getRemoteAccessPort()
     */
    public int getRemoteAccessPort() {
        return remoteAccessPort;
    }
    
    /**
	 * Returns the ContextManager
	 * 
	 * @return Returns the ContextManager
     * @see org.msda.components.interfaces.MSDAManagerInterface#getContextManager()
     */
    public MSDContextManager getContextManager() {
        return ctxManager;
    }
    
    /**
	 * Updates the cache of local components based on the received information about the other component.<br>
	 * In a specific case, the manager can activate the election of a bridge.
	 * 
	 * @param typeComponent		The component type (DEVICE, MANAGER, BRIDGE)
	 * @param info				The information about the component
	 *
	 * @see org.msda.components.interfaces.MSDAManagerInterface#updateInfoComponent(org.msda.msd.InformationComponent)
	 */
	public void updateInfoComponent(int typeComponent, InformationComponent info) {
		
	    switch (typeComponent)
		{
		case InfoType.Component.BRIDGE:
			localMgr.addBridge(info);
			localMgr.addComponent(info);
			break;
		case InfoType.Component.DEVICE:
		    if (!info.getAgreeBridge().equals(BigInteger.valueOf(0))) {
		        board.debug("Elect a component as a bridge");
		        this.daemon.sendActivateBridge (info);
		        this.localMgr.addBridge(info); 
		    }
			// the add component checks for duplicates...
			localMgr.addComponent(info);
			break;
		case InfoType.Component.MANAGER:
			localMgr.addManager(info);
			localMgr.addComponent(info);
			break;
		}
	}
    
	
	/**
	 * 
	 * @param typeComponent		The component type (DEVICE, MANAGER, BRIDGE)
	 * @param info				The information about the component
	 */
	public void removeInfoComponent(int typeComponent, InformationComponent info)
	{
	    switch (typeComponent)
		{
		case InfoType.Component.BRIDGE:
			localMgr.removeBridge(info);
			break;
		case InfoType.Component.DEVICE:
			localMgr.removeComponent(info);
			break;
		case InfoType.Component.MANAGER:
			localMgr.removeManager(info);
			break;
		}		
	}

	
    //
	// 
	//
    
    /**
     * Returns the ThreadDescriptionMgr.
     * 
     * @return Returns the ThreadDescriptionMgr.
     */
    public ThreadDescriptionMgr getThreadDescriptionManager() {
        return descManager;
    }
    
	
	/* RequestInterface interface methods */
	
	/**
     * Performs a service discovery.
     * Receives a message from the MSDService to perform a discovery request. First prepares 
     * the MSD request to be processed (put into the blocking queue of things to to)
     * Then wait for the pending request to be done and return the result. 
     * If TEST is true, return as soon as ONE (1) result is obtained, otherwise, wait for the
     * request to be done.
     * 
     * @param request	The request of the client.
     * 
     * @return	Returns all the MSD Descriptions that satisfies the MSD Request. 
     * @see org.msda.components.interfaces.RequestInterface#getServices(org.msda.msd.MSDRequest)
     */
	public MSDDescription[] getServices( MSDRequest request ) 
	{
		PendingRequest pending;
		ReqCreationInformation rci;
		Creator creator;
		long clientTimeout;
        PropagationInformation propagation;
        //
        int clientId = -1, requestId = -1;
	    //
	    ObjectLocalRequest object;
	    Vector results ;
    	long start, end;

        board.info("[MSDAManager] received an msda discovery request");

    	long processing_time_end, processing_time_start;
    	processing_time_start = System.currentTimeMillis();
	   
    	BigInteger bigClientId = request.getClientID();
    	if ( bigClientId!=null && bigClientId.intValue() < 0 )
    	    bigClientId = null;
    	
	    if ( bigClientId==null && !InfoType.TEST_SOAP_CALL ) {
	        
	        clientId = localCounterClient++;
		    requestId = localCounterRequest++;
	        
	        // First time the request is received, this is the creation of the request...
	        if ( request.getProcInfo()!=null ) {
	            rci = request.getProcInfo().getCrea();
	            creator = rci.getCreator();
	            if ( creator==null )
	                creator = new CreatorImpl();
	            creator.setNetID( this.netId );
	            creator.setMgrID( this.getMSDManagerID() );
	            
	            clientTimeout = rci.getLifetime().longValue();
	            if ( clientTimeout > 0 )
	                TimeoutGetServices = clientTimeout * 1000;
	        }
	        
	        board.info("TimeoutGetServices ="+TimeoutGetServices);
	        
	        request.setClientID( BigInteger.valueOf( clientId ) );
	        request.setReqID( BigInteger.valueOf( requestId ) );
	        
	        // create the pendingRequest here...
	        pending = new PendingRequest( request, TimeoutGetServices );
	        boolean exist = this.pendingRequestMgr.existPendingRequest( pending );
	        if ( !exist )
	            this.pendingRequestMgr.addPendingRequest( pending );
	        
	        
	        // Add the object to the queue, then the first available local thread will process it
	        object = new ObjectLocalRequest();
	        object.setRequest( request );
	        object.setClientTimeout( TimeoutGetServices );
	        object.setExistPendingRequest( exist );
	        
	        this.managerPool.getBlockingQueue().add( object );

	        board.info("[MSDAManager] request object in queue, wait for response");

	        try {
	            Thread.sleep(InfoType.WAIT_PENDING_REQUEST_CREATION);
		        board.info("[MSDAManager] wait ended");
	        } catch (InterruptedException e) {
		        board.info("[MSDAManager] wait interrupted");
	        }
                
	    } else {
	        
	        if ( !InfoType.TEST_SOAP_CALL ) {
		    	clientId = request.getClientID().intValue();
		    	requestId = request.getReqID().intValue();
			    pending = pendingRequestMgr.getPendingRequest( clientId, requestId );
	        }
	    }
	    	    
	    // in all cases, get all the current results...
	    results = new Vector();
	    
	    Vector tmp = null;
	    if ( InfoType.TEST_CTX )
	        tmp= new Vector();
	    
	    // If return when the first result is obtained, then I loop 
	    // and check the pending request each time until the timeout 
	    // expires or a result is found
	    start = System.currentTimeMillis();
	    
	    while(true) {
	        results = pending.getResults();
	        
	        if ( InfoType.TEST_CTX && !results.isEmpty() )
	            tmp.addAll( results );
	        board.info ("waiting for a result " + results.size());
	        
	        // I need to get the results each time...
	        end = System.currentTimeMillis() - start;               
	        if ( end >= TimeoutGetServices) {
	            break;
	        }
            // if context management, process context rules
            if ( InfoType.TEST_CTX && tmp!=null && !tmp.isEmpty() ) {
	            if ( request.getCtxRules()!=null && request.getCtxRules().getCRSvc()==null ) {
	                if (  tmp.size()==4 ) {
	                    results = tmp; 
	                    break;
	                }
	            } else {
	                results = tmp; 
	                break;
	            }
	        } else if ( results!=null && !results.isEmpty() ){
	            // leaving as soon as the first result is received
	            board.info ("MSDManager : found a result for " + clientId + " " + requestId);
	            break;
	        }
	        try {
	            Thread.sleep(2*InfoType.WAIT_PENDING_REQUEST_FIRST_RESULT);
	        } catch (InterruptedException e) {
	        }
	    }
    	board.info ("MSDManager : get services ended for  " + clientId + " " + requestId);

	    // if no result, return a dummy response
	    // TODO: check that this is the proper way...
	    if (results==null || results.isEmpty() ) {
	    	board.info ("but no result to return...");
	    	StringBuffer buf;
	    	UbisecID deviceId;
	    	Creator tmpCreator;
	    	CreationInformation creation;
	    	
	    	// create a result with InfoType.DUMMY_DESC_RUNNING_REQUEST
	        buf = new StringBuffer();
	        buf = buf.append( clientId ).append( "#").append( requestId );
	        
	        deviceId = new UbisecIDImpl();
	        deviceId.setName( buf.toString() );
	        
	        tmpCreator = new CreatorImpl();
	        tmpCreator.setDvcID( deviceId );
	        creation = new CreationInformationImpl();
	        creation.setCreator( tmpCreator );
	        
	        InfoType.DUMMY_DESC_RUNNING_REQUEST.setCreaInfo( creation );
	        results.add( InfoType.DUMMY_DESC_RUNNING_REQUEST );
	        
	        if ( InfoType.TEST_SOAP_CALL )
	            return (MSDDescription[])results.toArray( new MSDDescriptionImpl[0]);
	    }

	    if (this.isContextEnabled)
	    {
		    if ( request.getCtxRules()!=null ) 
		    {
		        if ( request.getCtxRules().getCRSvc()!=null ) {
		          
		        // Registration of the user and her context with the ContextManager	     
		        
//		        List list = request.getCtxInfo().getCIUser().getCITable();
//		        ContextElement elt = ContextUtil.getInstance().getContextElement( UserTemplate.DEVICE_NAME, list );
//		        String deviceName =  elt.getVal().getStr();
		        //print("deviceName is " + deviceName);
		        //ctxManager.addUser(deviceName, request.getCtxInfo().getCIUser(), request.getCtxRules().getCRUser());
		                
		        // ORI: TODO check if the update is necessary
		        // ORI: TODO there should be also a check to eliminate old service descriptions
		        
		        // Save services ?
		        
		        // update list of available services in the ContextManager     
		        //print("--------- save services -------");
		        //print("results is " + results.toString());
		        //print("results size " + results.size());
//		        for (int k = 0; k < results.size(); k++) {
//		            MSDDescription description = (MSDDescription)results.elementAt(k);
//		            if ( !InfoType.RUNNING_REQUEST.equals( description.getDescID().getName() ) 
//		                    && !InfoType.TIMEOUT_EXPIRED.equals( description.getDescID().getName() ) ) { 
//		                //print("service " + k + description.getServiceDefinition().getGeneric().getName());
//		                //print("service state " + description.getContextInformation().getServiceState());
//		                ServiceContext serviceCtx = new ServiceContext(description);	 
//		                //print("service context to save in the context manager " + serviceCtx.toString());
//		                //ctxManager.addService(serviceCtx);
//		            }
//		        }
		        
		        // perform matching between the user context and the contexts of the discovered services
		        Vector matchedResults = new Vector();	        
		        //print("------ match services and user -------");
		        for (int i = 0; i < results.size(); i++) {
		            MSDDescription description = (MSDDescription)results.elementAt(i);	            
		            if ( !InfoType.RUNNING_REQUEST.equals( description.getDescID().getName() ) 	                    
		                    && !InfoType.TIMEOUT_EXPIRED.equals( description.getDescID().getName() ) ) { 
		                //print("service " + i + description.getServiceDefinition().getGeneric().getName());
		                ContextInformationTable serviceCtx = description.getCtxInfo().getCISvc();
		                //print("invoke matcher");
		                ContextMatcher matcher = new ContextMatcher(request, serviceCtx);
		                //long time = System.currentTimeMillis();
		                if(matcher.matchUserService())
		                    matchedResults.add(description);
		                //long diff = System.currentTimeMillis() - time;
		                //board.info( Long.toString( diff ) );
		            }
		        }
		        board.info("matchedResults.size ="+matchedResults.size());
		        results = matchedResults;
		        }
		    }
	    }
	    
        //processing_time_end  = System.currentTimeMillis();
        //timelog.info (""+(processing_time_end-processing_time_start));
        //board.info ("msd processing time " + (processing_time_end-processing_time_start));
        
        // add results for future svc access 
    	board.info ("return all the results..." + results.size());
        for( int i=0; i<results.size(); i++) {
            MSDDescription desc = (MSDDescription)results.get(i);
            if ( !InfoType.RUNNING_REQUEST.equals( desc.getDescID().getName() ) && 
                 !InfoType.TIMEOUT_EXPIRED.equals( desc.getDescID().getName() ) ) {
                
                SvcAccessCacheObject accessObject = new SvcAccessCacheObject();
                accessObject.setPluginName( desc.getSvcDef().getOriginal().getProtocolName() );
                if ( desc.getPropaInfo()==null)
                    accessObject.setPropagation( null );
                else 
                    accessObject.setPropagation( desc.getPropaInfo().getHops() );
                this.svcAccCacheMgr.add( desc.getDescID().getName(), accessObject );
            }
        }
        
	    return (MSDDescription[])results.toArray( new MSDDescriptionImpl[0]);
	}
	
	
	/**
	 * Performs a remote service access 
	 *    - from the client to the plugin
	 *    - from the manager of the client's network
	 * 
	 * @param soapMessage	The access as a soap message
	 * @param descId		The MSD description ID of the service to access
	 * @param epService		The endPointReference of the service, i.e. its IP address in its network
	 * 
	 * @return 	The soap message response (as a String) of the service
	 * @see org.msda.components.interfaces.RequestInterface#serviceAccess(java.lang.String, int, java.lang.String)
	 */
	public String serviceAccess(String soapMessage, String descId, String epService) 
	{
		board.info( "Starting a serviceAccess... " + " descId " + descId );
		
		int clientID = localCounterRemote++;
		
		// search in the cache for service access...
		SvcAccessCacheObject cachedObject = this.svcAccCacheMgr.get( descId );
		if ( cachedObject==null ) {
		    board.info("ERROR cachedObject is null !");
		    return null;
		}
		
	    // <message >
	    //    <sender> ...  </sender>
        //    <type>RemoteAccess</type>
        //    <responseCode> clientid </responseCode>
        // ....
        // </message>
	    
	    MSDMessage message = new MSDMessageImpl();
	    message.setSender( this.managerId );
	    message.setType( BigInteger.valueOf( InfoType.Message.REMOTE_ACCESS ) );
	    message.setRespCode( BigInteger.valueOf( clientID ) );
	    
	    ServiceAccess svcAccess = new ServiceAccessImpl();
	    svcAccess.setSOAPMess( soapMessage );
	    svcAccess.setEPSvc( epService );
	    
	    //	  search the plugin name corresponding to this clientID
	    svcAccess.setPluginName( cachedObject.getPluginName() ); 
	    svcAccess.setSrcNetID( this.netId );
	    
	    ObjectServiceAccess object = new ObjectServiceAccess();
	    if ( cachedObject.getPropagation()!=null )
	        object.setHops( cachedObject.getPropagation() );
	    
	    MessageContent content = new MessageContentImpl();
	    content.setSvcAccess( svcAccess );
	    message.setContent( content );
	    
        object.setMessage( message );
        
        // Try to create the tunnel if the service is remote, 
        // in that goal, add the message to the queueRemoteAccess to notify a free thread
        this.managerPool.getBlockingQueue().add( object );
	  
        // wait for a response
        if ( object!=null ) {
        	
        	synchronized( object ) {
        	    try {
        	        object.wait( TimeoutServiceAccess );
        	    } catch (InterruptedException e) {
        	    }
        	}	        
            
            // return the response directly, or null if there was no response
            return object.getResponse();
        }
		return null;
	}

	/* CommunicatorCallback interface methods */
	
	/**
	 * Process an incoming MSDA description from the msd multicast group
	 * This method is called when a MSDA description has been received from the MSDA multicast group
	 * (i.e., forwarded by a MSDA bridge). This description should be added as a a remote service. 
	 * <p>
	 * 
	 * @see org.msda.components.interfaces.CommunicatorCallback#newDescription(org.msda.msddescription.MSDDescription)
	 *
	 */
	public void newDescription(MSDDescription description)
	{
	    if  ( description==null )
	        return;
	    
		// If same creator, no need to process this description
	    // Deals with the multicast of a local description
	    
	    // WARNING if mgr and bridge ? deactivated... (ADLC)
//		Creator creator = description.getCreationInformation().getCreator();
//		if (MSDUtil.compareID( creator.getManagerID(), this.managerId)) {
//			board.debug ("MSDManager::NewDescription : received a remote description. Same Creator, do nothing...");
//			return;
//		}
		
		// This is a remote description (from another msd manager), add it
		// The description will be added if it has not already been cached,
		// otherwise, the description will be updated
		boolean updated = descManager.updateService( description, false );
		if ( !updated )
			descManager.addService ( description, false );
	}

	/**
	 * Process an incoming MSD request (remote) from the MSD multicast group
	 * This method is called when a msd request has been received from the MSD multicast group 
	 * (i.e., forwarded by a MSD bridge). This request should be evaluated.
	 * <p>
	 * 
	 * @see org.msda.components.interfaces.CommunicatorCallback#newRequest(org.msda.msdrequest.MSDRequest)
	 */
	public void newRequest (MSDRequest request)
	{
		// If creator of the request, no need to process it (already pending locally)
		Creator creator = request.getProcInfo().getCrea().getCreator();
		if ( MSDUtil.getInstance().compareID( creator.getMgrID(), this.managerId ) ) {
			//board.debug ("MSDManager::newRequest : received a remote request. Same Creator, do nothing...");
			return;
		}
		
		ObjectRemoteRequest object = new ObjectRemoteRequest();
		object.setRequest( request );
		
		// Add the object to the queue, then a thread will process it
		this.managerPool.getBlockingQueue().add( object );
	}
	
	
	/**
	 * Receives an incoming MSD message from the MSD group to process
	 * 
	 * @param message : message received from the MSD group
	 */
	public void responseRequest( MSDMessage message ) {
		List hops;
		MSDDescription desc;
		int requestId;
		Param param;
		int clientId;
		boolean ok;
		
	    // The bridge removed the hop when the message comes from a network to the manager's network
	    // So, if the list of hops (in propagation information) is empty, then this manager is the 
		// destination manager, otherwise not the recipient and do nothing

		// Search in the propagation list
        if ( message.getBackwardPath()!=null ) {
            hops = message.getBackwardPath().getHops();
            if ( hops.isEmpty() ) {
            	
            	desc = message.getContent().getDesc();
            	requestId = message.getRespCode().intValue();
            	
            	// TODO : make sure this is the client ID (hardcoded value in responseRequest)
            	param = (Param)desc.getMetaInfo().getParams().get(0);
                clientId = Integer.parseInt( param.getValue() );
                //board.info("#1 MANAGER desc id ="+desc.getDescID().getName() + " "+System.currentTimeMillis());
            	ok = pendingRequestMgr.addResult( clientId, requestId, desc );
            	if (!ok)
            	{
            	    board.debug ( " MSDManager::responseRequest - PROBLEM: could not find the pending request for ( clientId " + clientId + " requestId " + requestId +" )" );
            	}
            	//board.info("#2 MANAGER desc id ="+desc.getDescID().getName() + " "+System.currentTimeMillis());
            } 
        }
	}
	
	/**
	 * Receives an incoming MSD message of type InformationComponent.<br>
	 * The MSD message contains the IP address and the port to process a remote access.
	 * 
	 * @param message received from the msd group
	 * 
	 *  @see org.msda.components.interfaces.CommunicatorCallback#getInformationComponents(org.msda.msd.MSDMessage)
	 */
    public void controlMessage( MSDMessage message ) {
    }
    
	/* PluginCallback interface methods */


    /**
	 * Registers a new service (in the MSD Manager) by providing its msd description
	 * 
	 * @param description : description received from the MSD group. May come from the MSD manager or from MSD bridges 
	 * @see org.msda.components.interfaces.PluginCallback#NewService(org.msda.MSDDescription.MSDDescription)
	 */	
	public void newService (MSDDescription description )
	{	  
		// Try to update the service if it has already been registered
	    boolean updated = descManager.updateService( description, true );
	    if ( !updated ) {
	        // add the new service
	        descManager.addService ( description, true );
	    }
	    
	    // TODO check if the update is necessary
	    // TODO there should be also a check to eliminate old service descriptions
	    // update list of available services in the ContextManager
	    if (this.isContextEnabled)
	    {
		    ServiceContext serviceCtx = new ServiceContext(description);
		    if (description.getCtxInfo()!=null)
		    {
		        ctxManager.addService(serviceCtx);    	    
		    }
	    }
		    
	}

	/**
	 * Cancels a service registerd (in the msd manager) by providing its unique id
	 * 
	 * @param id the unique id of the service to cancel
	 * @return true if service cancelled, false otherwise.
	 * @see org.msda.components.interfaces.PluginCallback#EndService(java.lang.String)
	 */
	public boolean endService ( String id  )
	{
		return descManager.removeService ( id, true );
	}

	/**
	 * Returns the MSD description associated with a local service by providing its unique ID
	 * <p>
	 * TODO check that this method is really necessary.
	 * 
	 * @param id the unique ID of the requested service
	 * @return the MSD description of the requested service
	 * @see org.msda.components.PluginCallback#GetLocalService(java.lang.String)
	 */
//	public MSDDescription getLocalService (String id)
//	{
//		return descManager.getService ( id, true);
//	}

	/**
	 * Returns a new MSD description, initialized with the creation information
	 * 
	 * @return the new msd description to be filed by the SD plugin
	 * @see org.msda.components.interfaces.PluginCallback#newMSDDescription()
	 */
	public MSDDescription newMSDDescription ()
	{
		CreationInformation newCreation;
		Calendar rightNow;
		Creator newCreator;
		MSDDescription newDescr;

		// create a new empty msd description with a unique id
		// WARNING: should create correctly otherwise it may not conform to the schema (all required elements should be there)
		newDescr = new MSDDescriptionImpl();
		newDescr.setDescID(newMSDDescriptionID());
		
		newCreation = new CreationInformationImpl();
		newCreator = new CreatorImpl();
		
		rightNow = Calendar.getInstance();
		
		newCreator.setDvcID(daemon.getStackId());				
		newCreator.setNetID(communicator.getNetworkID());				
		newCreator.setMgrID(managerId);	
		newCreation.setCreator(newCreator);
		newCreation.setCreaDate(rightNow);
		newCreation.setCreaTime(rightNow);
		newCreation.setUpdDate(rightNow);
		newCreation.setUpdTime(rightNow);
		newCreation.setUpdNb( BigInteger.valueOf(0));
		newCreation.setLifetime( BigInteger.valueOf(0));
		
		newDescr.setCreaInfo(newCreation);		
		
		return newDescr;
	}


	/**
	 * Starts a specific SD plugin that is dynamically loaded. 
	 * 
	 * @param pluginName 		the name of the plugin to load and start
	 * @param pluginCreator		the name of the plugin creator
	 * @param pluginClassName	the name of the java class to load
	 * @param args 				the arguments to start the plugin
	 * @param httpServerPort	server port for the plugin
	 * @param directMSDPort		server port for direct (socket) connection to the manager
	 * @return true if the plugin is started, false otherwise
	 *
	 * @see org.msda.components.interfaces.MSDAManagerInterface#startPlugin(java.lang.String, java.lang.String[])
	 */
	public boolean startPlugin (String pluginName, String pluginCreator, String pluginClassName, String[] args, int httpServerPort, int directMSDPort )
	{
		Class[] intfs;
		Class pluginClass;
		ClassLoader cl = ClassLoader.getSystemClassLoader() ; 
		PluginInterface tmpInterface;
		try {
			board.info ("* Loading the plugin with class name " + pluginClassName );
			pluginClass = cl.loadClass( pluginClassName ) ;
			intfs = pluginClass.getInterfaces();     // []
			for (int x=0; x<intfs.length; x++) 
			{
				board.info ( "has interfaces " + intfs[x].toString() );
				if (intfs[x].toString().equals("interface org.msda.components.interfaces.PluginInterface"))
				{
					board.info ( "correct interface, creating the plugin" );
					tmpInterface = (PluginInterface)pluginClass.newInstance();
					tmpInterface.setPluginInfo(pluginName, pluginCreator, httpServerPort, directMSDPort);
					for (int y=0; y<args.length; y+=2)
					{
						board.info ( "add params " + args[x] + " value " + args[x+1] );
						tmpInterface.setPluginParam(args[x], args[x+1]);
					}

					tmpInterface.registerManager ( this, this );
					pluginList.add (tmpInterface);
					
					if (isActive) {
						board.info ( "finally, activate plugin" );
						tmpInterface.activate();
					}
				}
			}
		}
		catch (ClassNotFoundException e) 
		{
		    board.debug( "MSDManager::startPlugin : class not found "+e.getMessage() );
			return false;
		}
		catch (InstantiationException e)
		{
		    board.debug( "MSDManager::startPlugin : could not instantiate "+e.getMessage() );
			return false;
		}
		catch (IllegalAccessException e)
		{
		    board.debug( "MSDManager::startPlugin : illegal access "+e.getMessage() );
			return false;
		}
		catch (MSDException e)
		{
		    board.debug( "MSDManager::startPlugin : ERROR when registering plugin " + e.getMessage() );
			return false;
		}
		return true;
	}

	/**
	 * Stops a specific SD plugin
	 * 
	 * @return true if stopped, false otherwise 
	 *
	 * @see org.msda.components.interfaces.MSDAManagerInterface#stopPlugin()
	 */
	public boolean stopPlugin ()
	{
		return true;
	}

	/**
	 * <p>
	 * TODO : here, needs to decide if transformers are only to be dealt with in plugins..
	 * 
	 * @see org.msda.components.interfaces.MSDAManagerInterface#startTransformer()
	 */
	public boolean startTransformer ()
	{
		return true;
	}

	/**
	 * <p>
	 * TODO : here, needs to decide if transformers are only to be dealt with in plugins...
	 * 
	 * @see org.msda.components.interfaces.MSDAManagerInterface#stopTransformer()
	 */
	public boolean stopTransformer ()
	{
		return true;
	}


	/**
	 * Returns the creator id for this msd manager.
	 * The creator id is a unique id within the heterogeneous system that allows to identify 
	 * the msd descriptions and requests in the system. It is the association of the device id, 
	 * the network id, and the msd manager id.
	 * 
	 * @return the creator id
	 */
	public Creator getCreator ()
	{
		Creator newCreator = new CreatorImpl();
		newCreator.setDvcID(daemon.getStackId());				
		newCreator.setNetID(communicator.getNetworkID());				
		newCreator.setMgrID(managerId);	
		return newCreator;
	}

	/**
	 * Creates a string containing a unique ID based on the time. 
	 * <p>
	 * TODO: this unique id may in fact not be unique system wide. The creator id and a local counter value 
	 * should be added to really have unique IDs.
	 * 
	 * @return the unique ID created
	 */
	private UbisecID newMSDDescriptionID()
	{
		return daemon.getUbisecID();
	}


	/**
	 * Returns true if the manager is active, false otherwise.
	 * 
     * @return Returns the isActive.
     */
    public boolean isActive() {
        return isActive;
    }

	
    /**
     * Calls the PendingRequestMgr.addResult method.
     * 
     * @param clientId
     * @param requestId
     * @param result
     * @return
     */
    public boolean addResult ( int clientId, int requestId, MSDDescription result ) {
        return pendingRequestMgr.addResult( clientId, requestId, result );
    }
    
	//
	// Needed for the GUI
	//
	
    /**
     * Registers the ManagerGUI.
     */
	public void registerGUI( GUINotification gui ) {
	    this.descManager.setGUINotification( gui );
	}
	
	
	/**
	 * Gets the list of activated plugins for this manager. 
	 * 
	 * @return Returns a vector of PluginInterface.
	 */
	public Vector getPlugins() {
	    
	    Vector v = new Vector();
	    if ( managerPool.getTaskPlugins()!=null )
	        for( int i=0; i<managerPool.getTaskPlugins().length; i++) {
	            v.add( managerPool.getTaskPlugins()[i].getPlugin() );
	        }
	    return v;
	}
	
	/**
	 * Returns the cache of local descriptions.
	 * 
	 * @return Returns the cache of local descriptions.
	 */
	public Vector getLocalDescriptions() {
	    return this.descManager.getLocalDescriptions();
	}
	
	/**
	 * Returns the cache of remote descriptions.
	 * 
	 * @return Returns the cache of remote descriptions.
	 */
	public Vector getRemoteDescriptions() {
	    return this.descManager.getRemoteDescriptions();
	}
	
	/**
	 * Returns the cache of neighbour components
	 * 
     * @return Returns the localMgr.
     */
    public LocalComponentsMgr getLocalMgr() {
        return localMgr;
    }
}
