/*
 * Created on Apr 22, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.manager;


import java.util.HashMap;
import java.util.Map;

import org.msda.components.interfaces.CommunicatorInterface;
import org.msda.components.interfaces.PluginInterface;
import org.msda.components.manager.client.BlockingQueue;
import org.msda.components.manager.object.ObjectPlugin;
import org.msda.components.manager.task.TaskPlugin;
import org.msda.components.manager.task.ThreadTask;
import org.msda.components.utils.LocalComponentsMgr;
import org.msda.msd.MSDRequest;
import org.msda.util.InfoType;

/**
 * This class allows the pool of threads to be managed by the msd manager.
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 */
public class MSDManagerPool {
    private ThreadDescriptionMgr descManager;
    private CommunicatorInterface communicator;
    private LocalComponentsMgr localMgr;

    private int remoteAccessPort;
 
	private ThreadRemoteAccessListener threadRemoteAccessListener;
	
	// Pool of threads
	private ThreadTask[] pool;
	private BlockingQueue blockingQueue;
	
	// Pool to manage plugins
	private TaskPlugin[] taskPlugins;

	private Map pluginIndex;
	private String networkName;
	
	private MSDManager manager;
	
	/**
	 * Constructor.
	 * 
	 * @param manager
	 * @param communicator
	 */
	public MSDManagerPool( MSDManager manager, CommunicatorInterface communicator ) {
	
	    // Initializes the pool of threads
		long start = System.currentTimeMillis();
		
		this.manager = manager;
		this.networkName = manager.getNetworkName();
		this.descManager = manager.getThreadDescriptionManager();
		this.communicator = communicator;
		this.localMgr = manager.getLocalMgr();
		this.remoteAccessPort = manager.getRemoteAccessPort();
	    this.pluginIndex = new HashMap();
		
	    blockingQueue = new BlockingQueue();
	    pool = new ThreadTask[ InfoType.DEFAULT_THREADS_MANAGER_POOL ];
	    for( int i=0; i<this.pool.length; i++)
	        pool[i] = new ThreadTask( this );

		this.threadRemoteAccessListener = new ThreadRemoteAccessListener( remoteAccessPort, this );
	}

	/**
	 * @see java.lang.Object#finalize()
	 */
	public void finalize() throws Throwable
	{
		try 
		{
			this.networkName = null;
			this.descManager = null;
			this.communicator = null;
			this.localMgr = null;
		    this.pluginIndex = null;
			this.threadRemoteAccessListener.close();
			this.threadRemoteAccessListener = null;
			
			this.blockingQueue.close();
			this.blockingQueue = null;
			
			for( int i=0; i<this.pool.length; i++) {
		        pool[i].terminate();
		    }
			pool = null;
		}
		catch (Exception e) 
		{
			
		}
		finally
		{
			super.finalize();
		}
	}
	
	/**
	 * Returns the network name.
	 * 
	 * @return Returns the network name.
	 */
	public String getNetworkName() {
		return this.networkName;
	}
	
	/**
	 * Returns the remote access port.
	 * 
	 * @return Returns the remote access port.
	 */
	public int getRemoteAccessPort() {
		return remoteAccessPort;
	}

    /**
     * Returns the ThreadDescriptionMgr.
     * 
     * @return Returns the ThreadDescriptionMgr.
     */
    public ThreadDescriptionMgr getDescManager() {
        return descManager;
    }
    
    /**
     * Returns the communicator.
     * 
     * @return Returns the communicator.
     */
    public CommunicatorInterface getCommunicator() {
        return communicator;
    }
    
    /**
     * Returns the LocalComponentsMgr.
     * 
     * @return Returns the LocalComponentsMgr.
     */
    public LocalComponentsMgr getLocalMgr() {
    	return this.localMgr;
    }
    
    /**
     * Returns the pluginIndex.
     * 
     * @return Returns the pluginIndex.
     */
    public Map getPluginIndex() {
        return pluginIndex;
    }
    
    /**
     * Returns the blockingQueue.
     * 
     * @return Returns the blockingQueue.
     */
    public BlockingQueue getBlockingQueue() {
        return blockingQueue;
    }
    
    /**
     * Returns a reference to the MSD manager
     * 
     * @return	Returns a reference to the MSD manager
     */
    public MSDManager getMSDManager() {
        return manager;
    }
    
    
    /**
     * Adds a new task to manage a plugin in the pool
     * 
     * @param plugin
     */
    public void createNewTaskPlugin( PluginInterface plugin ) {
        
        int index;
        
        if ( taskPlugins==null ) {
            taskPlugins = new TaskPlugin[1];
            taskPlugins[0] = new TaskPlugin( plugin, this );
            index = 0;
            
        } else {
            //Increase the pool
            TaskPlugin[] tmp = taskPlugins;
            taskPlugins = new TaskPlugin[ tmp.length+1 ];
            taskPlugins[tmp.length] = new TaskPlugin( plugin, this );
            System.arraycopy( tmp, 0, taskPlugins, 0, tmp.length );
            index = tmp.length;
        }
        
        // add to the pluginIndex
        this.pluginIndex.put( plugin.getName(), new Integer( index ) );
    }
    
    /**
     * Returns the TaskPlugin[]
     * 
     * @return Returns the TaskPlugin[]
     */
    public TaskPlugin[] getTaskPlugins() {
        return taskPlugins;
    }
    
    /**
	  * Evaluates an MSD request against local SD plugins depending on the request mode.
	  * This method forwards it to the different SD plugins for evaluation. 
	  * <p>
	  * 
	  * @param request the msd request to be evaluated
	  * @param remote true if the request is remote, false otherwise
	  */
	 public void evaluateRequest ( MSDRequest request, boolean remote ) {
	     
		 if (taskPlugins==null)
			 return;
		 
		 // Contact the relevant plugins (only if the plugin is in passive mode)
		 PluginInterface tmpInterface;
		 
		 for (int i=0; i<taskPlugins.length; i++) {
		     tmpInterface = taskPlugins[i].getPlugin();
		     if ( !tmpInterface.doNotify() ) {
		         // sends the request to the plugin and activates the TaskPlugin
		         ObjectPlugin object = new ObjectPlugin();
		         object.setRequest( request );
		         object.setRemote( remote );		
		         object.setIndex( i );
		         getBlockingQueue().add( object );
		     }
		 }
	 }
}
