/*
 * Created on Apr 19, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.manager;

import java.util.Calendar;
import java.util.Vector;

import org.msda.msd.MSDDescription;
import org.msda.msd.MSDRequest;
import org.msda.util.InfoType;
import org.msda.util.MSDUtil;

/**
 * This class is used to store information about pending requests.<br> 
 * Especially, it contains : <br>
 *   - the request itself, <br>
 *   - the currently available responses to this request, <br> 
 *   - the request id assigned by the msd manager, <br> 
 *   - the client timeout to perform the request, <br>
 *   - and its validity time (2* client timeout)  
 * 
 * @author raverdy, delachap - INRIA ARLES
 */
public class PendingRequest {

    private MSDRequest request;
	private Vector resultMSDDescriptionList;
	
	// use to not put in cache twice the same description 
	private Vector resultSendList;
	
	private int requestId;
	
	private long clientTimeout;
	private Calendar startTime;
	private Calendar validityTime;
	private Calendar timeoutTime;
	
	/**
	 * Creates a pending request data structure with a specific lifetime.
	 * 
	 * @param request the msd request
	 * @param lifetime the lifetime of the request
	 */
	public PendingRequest (MSDRequest request, long clientTimeout)
	{
		this.request = request;
		this.clientTimeout = clientTimeout;
		
		this.requestId = request.getReqID().intValue();
		this.resultMSDDescriptionList = new Vector();
		this.resultSendList = new Vector();
		
		startTime = Calendar.getInstance();
		timeoutTime = Calendar.getInstance();
		timeoutTime.add( Calendar.MILLISECOND,  (int)clientTimeout );
		
		int twoTimeout = 2 * (int)clientTimeout;
		validityTime = Calendar.getInstance();
		validityTime.add( Calendar.MILLISECOND, twoTimeout );
	}
	
    /**
     * Returns the MSD request
     * 
     * @return Returns the MSD request.
     */
    public MSDRequest getRequest() {
        return request;
    }
    
    /**
     * Returns the request id.
     * 
     * @return Returns the requestId.
     */
    public int getRequestId() {
        return requestId;
    }
    
    /**
	 * Adds a MSD description as a result to the request.
	 * TODO - modify the test, compare the sent description, not only the object... 
	 * 
	 * @param desc - the result
	 * @return true if the MSDDescription has been added, false otherwise
	 */
	public boolean addResult( MSDDescription desc ) {
	    
	    // Verify the timeout
	    if ( !checkTimeout() ) 
	    {
	    	//board.info ("PendingRequest::addResult - cannot add timeout expired...");
	        return false;
	    }
	    // Verify this result has not already been sent.
	   if ( !MSDUtil.getInstance().containsDescription( this.resultSendList, desc ) && 
	            !MSDUtil.getInstance().containsDescription( this.resultMSDDescriptionList, desc ) ) {
	      
	      synchronized(this.resultMSDDescriptionList) {
	          this.resultMSDDescriptionList.add( desc );
	      }
	      //board.info ("PendingRequest:: add ok "+System.currentTimeMillis() );
	      return true;
	    }
	    return false;
	}
    
    /**
     * Gets the current gathered results.
     * 
     * @return 	vector of MSDDescription
     */
    public synchronized Vector getResults() {
	    
        boolean last = false;
        if ( !checkTimeout() ) 
	        last = true;
        
        Vector results = new Vector();
	    for(int i=0; i<resultMSDDescriptionList.size(); i++) {
	        MSDDescription desc = (MSDDescription)this.resultMSDDescriptionList.get(i);
		    this.resultSendList.add( desc );
	        results.add( desc );
	    }
	 
	    if ( last ) {
	    	//board.info ("PendingRequest::getResults - adding timeout expired result...");
	        results.add( InfoType.DESC_TIMEOUT_EXPIRED );
	    }
	    
	    this.resultMSDDescriptionList.clear();
	    return results;
	}
    
    /**
     * Checks if the pending request is still valid, i.e. 2*timeout has expired or not.
     * 
     * @return true if it has not expired, false otherwise
     */
    public boolean checkValidity() {
     
        Calendar currentTime = Calendar.getInstance();
        if ( currentTime.before( validityTime ) )
            return true;
        return false;
    }
    
    /**
     * Checks if the timeout has expired or not.
     * 
     * @return true if it has not expired, false otherwise
     */
    private boolean checkTimeout() {
	    Calendar currentTime = Calendar.getInstance();
	    if ( currentTime.before( timeoutTime ) )
	        return true;
	    return false;
	}
}
