/*
 * Created on Jul 5, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.manager;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.msda.msd.CreationInformation;
import org.msda.msd.Creator;
import org.msda.msd.MSDDescription;
import org.msda.msd.MSDRequest;
import org.msda.msd.ReqCreationInformation;
import org.msda.msd.ReqProcessingInformation;
import org.msda.msd.UbisecID;
import org.msda.msd.impl.CreationInformationImpl;
import org.msda.msd.impl.CreatorImpl;
import org.msda.msd.impl.UbisecIDImpl;
import org.msda.util.MSDUtil;

/**
 * Class to manage the PendingRequests
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 *
 */
public class PendingRequestMgr {

    /**
     * Map composed of ( clientId, Vector of PendingRequests )
     */
    private Map localRequests;
    
    
    public PendingRequestMgr() 
    {
        localRequests = new HashMap();
        Collections.synchronizedMap( localRequests );
    }
    
    /**
     * Returns the set of PendingRequests corresponding to the same clientId.
     * 
     * @param clientId
     * @return Vector of PendingRequests
     */
	public synchronized Vector getPendingRequests( int clientId ) 
	{
		PendingRequest pending;
	    Collection c = (Collection)localRequests.get( new Integer( clientId ) );
	    if ( c==null )
	        return (Vector)c;
	    
	    // this method is called by both addResult and existPendingRequest
	    // check here if the PendingRequests are still valid
	    for (Iterator iter = c.iterator(); iter.hasNext();) 
	    {
            pending = (PendingRequest) iter.next();
            if ( !pending.checkValidity() ) 
            {
                iter.remove();
                //board.info("PendingRequestMgr (TO DELETE)- remove a pending request...");
            }
        }
	    
	    Vector v = (Vector)c;
	    localRequests.put( new Integer( clientId ), v );
	    
	    return v;
	}
	
	/**
     * Gets the PendingRequest corresponding to the clientId and RequestId as input parameters
     * 
     * @param clientId
     * @param requestId 
     * @return The corresponding PendingRequest
     */
	public synchronized PendingRequest getPendingRequest( int clientId, int requestId) 
	{
	    PendingRequest tmpRequest = null;
	    
	    Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null )
	        return null;
	    
	    for (int x=0; x<pendingRequests.size(); x++)
		{
	        tmpRequest = (PendingRequest)pendingRequests.get(x);
	        if ( tmpRequest.getRequestId()==requestId )
	            break;
		}
	    return tmpRequest;
	}
	
	
	/**
	 * Verifies if a PendingRequest already exists or not.
	 * 
	 * @param request	the PendingRequest
	 * @return true if the PendingRequest already exist, false otherwise
	 */
	public boolean existPendingRequest( PendingRequest request ) {
	    
	    int clientId = request.getRequest().getClientID().intValue();
	    Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null )
	        pendingRequests = new Vector();
	    
	    // Verify if the same Pending Request already exists  
	    PendingRequest tmpRequest;
	    MSDRequest msdRequest1, msdRequest2;
	    ReqProcessingInformation proc1, proc2;
	    Creator creator1, creator2;
	    for (int x=0; x<pendingRequests.size();x++)
		{
	        tmpRequest = (PendingRequest)pendingRequests.get(x);
			
	        msdRequest1 = tmpRequest.getRequest();
	        msdRequest2 = request.getRequest();
	        
	        // Two PendingRequests are the same if their clientID and creator.managerID
	        // are equals, because clientID is an identification local to a MSDManager
	        if ( msdRequest1.getClientID().equals( msdRequest2.getClientID() ) ) {
	            
	            proc1 = msdRequest1.getProcInfo();
	            proc2 = msdRequest2.getProcInfo();
	            
	            creator1 = proc1.getCrea().getCreator();
	            creator2 = proc2.getCrea().getCreator();
	            
	            UbisecID id1 = creator1.getMgrID();
	            UbisecID id2 = creator2.getMgrID();
	            if ( MSDUtil.getInstance().compareID( id1, id2 ) )
	                return true;
	        }
		}
	    return false;
	}
    	
	/**
	 * Adds a Pending Request to the Vector
	 * 
	 * @param request The PendingRequest to add
	 */
	public void addPendingRequest( PendingRequest request ) {
	 
	    int clientId = request.getRequest().getClientID().intValue();
	    Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null )
	        pendingRequests = new Vector();
	    
	    pendingRequests.add( request );
	    localRequests.put( new Integer( clientId ), pendingRequests );
	}
	
	/**
	 * Adds a result to a client's pending request
	 * 
	 * @param requestId id of the msd request that the result belongs to
	 * @param result the msd description that fits the request
	 * @return true if the result has been added, false otherwise
	 */
	public boolean addResult ( int clientId, int requestId, MSDDescription result )
	{
	    Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null ) {
	        //board.info("PendingRequestMgr :: pendingRequests is null");
	        return false;
	    }
	    
	    PendingRequest tmpRequest;
		for (int x=0; x<pendingRequests.size();x++)
		{
			tmpRequest = (PendingRequest)pendingRequests.get(x);
			if (tmpRequest.getRequestId() == requestId)
			{
			    // set the deviceID with clientId#requestId
			    StringBuffer buf = new StringBuffer();
			    buf = buf.append(clientId).append('#').append(requestId);
			    UbisecID deviceID = new UbisecIDImpl();
			    deviceID.setName( buf.toString() );
			    
			    CreationInformation creation = result.getCreaInfo();
			    if ( creation==null )
			    {
			        creation = new CreationInformationImpl();
			        result.setCreaInfo(creation);
			    }
			    Creator creator = creation.getCreator();
			    if ( creator==null )
			    {
			        creator = new CreatorImpl();
			        creation.setCreator(creator);
			    }
			    creator.setDvcID( deviceID );			    
				return tmpRequest.addResult( result );
			}
		}
		//board.info("PendingRequestMgr :: not found");
		return false;
	}
    
    /**
	 * If the client is a remote client, this value is the one given by its local network
     * 
     * @see ClientEntry#getSrcNetworkID()
     * 
     * @return the networkID of the MSDManager to which the client is connected
	 */
	public UbisecID getSrcNetworkID( int clientId ) {
	    
	    Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null )
	        return null;
	    
	    if ( pendingRequests.isEmpty() )
	        return null;
	    
	    MSDRequest request = ((PendingRequest)pendingRequests.get(0)).getRequest();
	    ReqCreationInformation creation = request.getProcInfo().getCrea();
	    
	    // This information is the MSDRequest.creator.NetworkID of any of the PendingRequests
	    return creation.getCreator().getNetID();
	}
	
    /**
     * If the client is a remote client, this value is the one given by its local manager.<br>
     * 
     * @see ClientEntry#getCreatorManagerID()
     * 
     * @return the ManagerID of the MSDManager to which the client is connected 
     */
    public UbisecID getCreatorManagerID( int clientId ) {
        
        Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null )
	        return null;
	    
	    if ( pendingRequests.isEmpty() )
	        return null;
	    
	    MSDRequest request = ((PendingRequest)pendingRequests.get(0)).getRequest();
	    ReqCreationInformation creation = request.getProcInfo().getCrea();
	        
        // This information is the MSDRequest.creator.ManagerID of any of the PendingRequests
        return creation.getCreator().getMgrID();
    }
    
    /**
     * If the client is a remote client, this value is the one given by its local manager
     * 
     * @see ClientEntry#getCreatorClientID()
     * 
     * @return the clientId defined by the MSDManager to which the client is connected 
     */
    public int getCreatorClientID( int clientId ) {
        
        Vector pendingRequests = getPendingRequests( clientId );
	    if ( pendingRequests==null )
	        return -1;
	    
	    if ( pendingRequests.isEmpty() )
	        return -1;
        
	    MSDRequest request = ((PendingRequest)pendingRequests.get(0)).getRequest();
	    
        //This information is the MSDRequest.clientID of any of the PendingRequests
        return request.getClientID().intValue();        
    }
}
