/*
 * Created on May 11, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.manager;

import java.math.BigInteger;
import java.util.Calendar;
import java.util.Vector;

import org.msda.components.manager.client.BlockingQueue;
import org.msda.components.manager.object.Object4Queue;
import org.msda.components.manager.object.ObjectRemoteRequest;
import org.msda.framework.FrameworkOutput;
import org.msda.gui.GUINotification;
import org.msda.msd.CreationInformation;
import org.msda.msd.MSDDescription;
import org.msda.msd.MSDRequest;
import org.msda.msd.ServiceDefinition;
import org.msda.util.MSDUtil;

/**
 * Manages the cache of local or remote descriptions.
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 */
public class ThreadDescriptionMgr extends Thread {
    
    /** Default timeout (20s) between each loop */
    private static final int BEAT_TIMEOUT = 20000;
    
    //DEFAULT LIFETIME - 1 min
    /** Default lifetime (60s) sets to the descriptions */
    private static final BigInteger DEFAULT_LIFETIME = BigInteger.valueOf( 60 );
    
	private Vector localMSDDescriptionList;
	private Vector remoteMSDDescriptionList;
	boolean doBeatLoop;
	private FrameworkOutput board;
	private GUINotification gui;
	
	/**
	 * Creates the description manager
	 * 
	 * @param board
	 */
	public ThreadDescriptionMgr ( ) {
		this.board = org.msda.framework.FrameworkMain.msdaOutput;
	    this.doBeatLoop = true;
		this.remoteMSDDescriptionList = new Vector();
		this.localMSDDescriptionList = new Vector();
		
		this.start();
	}


	/**
	 * Terminates the description manager. 
	 * The loop thread should be stopped. Vectors will be freed automatically.
	 *
	 */
	public void terminate () 
	{
		doBeatLoop = false;
	}

	/**
	 * Loop that checks for old descriptions to remove <br> 
	 * 	- in the cache of local descriptions <br>
	 *  - in the cache of remote descriptions  <br>
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run () 
	{	
	    while ( doBeatLoop ) {
	        
	        try {				
	            MSDDescription desc;
	            // checks the lifetime of the local descriptions
	            for (int i=0; i<localMSDDescriptionList.size(); ) {
	                desc = (MSDDescription)localMSDDescriptionList.get(i);
	                if ( checkLifetime( desc ) == false ) 
	                    localMSDDescriptionList.remove(i);
	                else 
	                    i++;
	            }
	            
	            // checks the lifetime of the remote descriptions
	            for (int i=0; i<remoteMSDDescriptionList.size(); ) {
	                desc = (MSDDescription)remoteMSDDescriptionList.get(i);
	                if ( checkLifetime( desc ) == false ) 
	                    remoteMSDDescriptionList.remove(i);
	                else
	                    i++;
	            }
	            
	            synchronized( this ) {
	                wait( ThreadDescriptionMgr.BEAT_TIMEOUT );
	            }
	            
	        } catch ( InterruptedException e ) {
	        }
	    }
	}

	/**
	 * Returns the number of either local or remote service descriptions currently stored by the MSD manager
	 * 
	 * @param isLocal true if local, false if remote
	 * @return size of local/remote descriptions stored
	 */
	public int getSize (boolean isLocal)
	{
		if (isLocal) 
			return localMSDDescriptionList.size();
		
		return remoteMSDDescriptionList.size();
	}

	/**
	 * Returns the Xth service description in either the local or remote service description lists.
	 * 
	 * @param x the number/position of the service to be returned (should be between 0 and (size-1)
	 * @param isLocal whether to get the service description in the local or remote list
	 * @return the msd description of the service, null if X is outside the bounds of the list
	 */
	public MSDDescription getService (int x, boolean isLocal)
	{
		if (isLocal) 
			return (MSDDescription)localMSDDescriptionList.get(x);
		
		return (MSDDescription)remoteMSDDescriptionList.get(x);
	}

	/**
	 * Returns the service description for service "id" in either the local or remote service description lists.
	 * 
	 * @param id the id of the service to be returned
	 * @param isLocal whether to look in the local or remote lists
	 * @return the msd description of the service, null if the id could not be found
	 */
	synchronized public MSDDescription getService (String id, boolean isLocal)
	{
		MSDDescription tmpDescr;
		//
		if (isLocal) {
			for (int x=0; x<localMSDDescriptionList.size(); x++)
			{
				tmpDescr = (MSDDescription)localMSDDescriptionList.get(x);
				if (tmpDescr.getDescID().equals( id ))
				{
					return tmpDescr;
				}
			}
		}
		else {
			for (int x=0; x<remoteMSDDescriptionList.size(); x++)
			{
				tmpDescr = (MSDDescription)remoteMSDDescriptionList.get(x);
				if (tmpDescr.getDescID().equals( id ))
				{
					return tmpDescr;
				}
			}
		}
		//
		return null;
	}

	/**
	 * Adds a service (local from a plugin, or remote from the msd group) to the
	 * manager's lists of msd descriptions. This method is synchronized to prevent inconsistencies if 
	 * multiple updates (add/delete) are received.
	 * 
	 * @param mdescr the msd description to add
	 * @param isLocal whether to add the msd description in the local or remote list
	 * @return returns true id the service was added, false if a problem occured (e.g., a local description could not be sent on the msd group)
	 */		
	synchronized public boolean addService ( MSDDescription mdescr, boolean isLocal )
	{
//	    if ( InfoType.TEST_PARTIAL_RESULTS ) // do nothing
//	        return false;
	    
	    if (isLocal)
	    {   
	        // the msd description is added to the local list
	        
	        checkLifetime( mdescr );
	        localMSDDescriptionList.add( mdescr );
	        board.info("----------------------------------");
	        board.info("MSDManager : add a local service "+mdescr.getDescID().getName() );
	        board.info("----------------------------------");
	        
	        if ( gui!=null )
	            gui.addDescription( true );
	    }
	    else	
		{
			// the msd description is added to the remote list
			
		    checkLifetime( mdescr );
			remoteMSDDescriptionList.add( mdescr );
			board.info("----------------------------------");
			board.info("MSDManager : add a remote service "+mdescr.getDescID().getName());
			board.info("----------------------------------");
			
			if ( gui!=null )
			    gui.addDescription( false );
		}
		return true;
	}

	/**
	 * Removes a service from the local/remote service lists based on its id
	 * 
	 * @param id the id of the service to delete
	 * @param isLocal whether to look in the local or remote service description lists
	 * @return true if service was removed, false if a problem occured (service not there)
	 */
	synchronized public boolean removeService ( String id, boolean isLocal )
	{
		boolean res = false;
		MSDDescription tmpDescr;
		//
		if (isLocal)
		{
			for (int x=0; x<localMSDDescriptionList.size(); x++)
			{
				tmpDescr = (MSDDescription)localMSDDescriptionList.get(x);
				if (tmpDescr.getDescID().equals( id ))
				{
				    board.info("----------------------------------");
				    board.info("MSDManager : removed a local service "+id);
				    board.info("----------------------------------");
					localMSDDescriptionList.remove(x);
					x--;
					res = true;
					
					if ( gui!=null )
					    gui.removeDescription( true );
				}
			}
		}
		else 
		{
			for (int x=0; x<remoteMSDDescriptionList.size(); x++)
			{
				tmpDescr = (MSDDescription)remoteMSDDescriptionList.get(x);
				if (tmpDescr.getDescID().equals( id ))
				{
				    board.info("----------------------------------");
				    board.info("MSDManager : removed a remote service "+id);
				    board.info("----------------------------------");
					remoteMSDDescriptionList.remove(x);
					x--;
					res = true;
					
					if ( gui!=null )
					    gui.removeDescription( false );
				}
			}
		}
		return res;
	}	
	
	
	/**
	 * Updates a service from its local/remote service lists
	 * 
	 * @param id the id of the service to update
	 * @param isLocal whether to look in the local or remote service description lists
	 * @return true if service was updated, false if a problem occured (service not there)
	 */
	synchronized public boolean updateService( MSDDescription description, boolean isLocal ) {
	    
	    MSDDescription tmpDescr;
	    if (isLocal)
		{
			for (int x=0; x<localMSDDescriptionList.size(); x++)
			{
				tmpDescr = (MSDDescription)localMSDDescriptionList.get(x);
				
				// First, verify the services could be the same...
				ServiceDefinition sd1 = description.getSvcDef();
				ServiceDefinition sd2 = tmpDescr.getSvcDef();
				if ( sd1.getGeneric().getName().equals( sd2.getGeneric().getName() ) ) {
					if ( !MSDUtil.getInstance().matchingDescription( description, tmpDescr ) )
					{
					    board.info("----------------------------------");
					    board.info("MSDManager : update a local service :"+description.getDescID().getName());
					    board.info("----------------------------------");
					    checkLifetime( description );
					    localMSDDescriptionList.set( x, description);
						return true;
					}
				} else 
				    return false;
			}
		}
		else 
		{
		    for (int x=0; x<remoteMSDDescriptionList.size(); x++)
			{
				tmpDescr = (MSDDescription)remoteMSDDescriptionList.get(x);
				
				// First, verify the services could be the same...
				ServiceDefinition sd1 = description.getSvcDef();
				ServiceDefinition sd2 = tmpDescr.getSvcDef();
				if ( sd1.getGeneric().getName().equals( sd2.getGeneric().getName() ) ) {
				    if ( !MSDUtil.getInstance().matchingDescription( description, tmpDescr ) )
				    {
				        board.info("----------------------------------");
				        board.info("MSDManager : update a remote service :"+description.getDescID().getName());
				        board.info("----------------------------------");
				        checkLifetime( description );
				        remoteMSDDescriptionList.set( x, description);
				        return true;
				    }
				} else 
				    return false;
			}
		}
	    return false;
	}
		

	/**
	 * Evaluates the msd request with the cache descriptions (local or remote).
	 * 
	 * @param mgr		MSDManager
	 * @param request	The request
	 * @param isLocal 	true to search in the cache of local descriptions, false in the cache of remote descriptions 
	 * @param forRemote	true if the request is remote, false if it is local
	 * 
	 * @return true if sth has been found, false otherwise
	 */
	synchronized public boolean evaluateRequest ( MSDManager mgr, BlockingQueue blockingQueue, MSDRequest request, boolean isLocal, boolean forRemote )
	{
		MSDDescription tmpDescription;
		boolean found = false;
	
		if (isLocal)
		{
		    for (int x=0; x<localMSDDescriptionList.size(); x++)
		    {
		        tmpDescription = (MSDDescription)localMSDDescriptionList.get(x);
		        
		        if ( MSDUtil.getInstance().matchingRequest ( request.getDescInfo().getService(), tmpDescription, request ))		        {
		            board.info("----------------------------------");
		            board.info("MSDManager : find a local service :"+tmpDescription.getDescID().getName());
		            board.info("----------------------------------");
		            found = true;
		            
					if ( !forRemote ) {
					    // case of a local result
					    mgr.addResult( request.getClientID().intValue(), request.getReqID().intValue(), tmpDescription);
					    
					} else {
					    // case of a remote result
					    ObjectRemoteRequest object = new ObjectRemoteRequest();
					    object.setRequestId( request.getReqID().intValue() );
					    object.setDescription( tmpDescription );
					    object.setRequest( request );
					    object.setType( Object4Queue.REMOTE_REQUEST4RESULT );
					    blockingQueue.add( object );
					}
		        }
		    }
		}
		else 
		{
		    for (int x=0; x<remoteMSDDescriptionList.size(); x++)
			{
		        tmpDescription = (MSDDescription)remoteMSDDescriptionList.get(x);
		        if ( MSDUtil.getInstance().matchingRequest ( request.getDescInfo().getService(), tmpDescription, request ))
		        {
		            board.info("----------------------------------");
		            board.info("MSDManager : find a remote service :"+tmpDescription.getDescID().getName());
		            board.info("----------------------------------");
		            found = true;
		            
		            if ( !forRemote ) {
		                // case of a local result
		                mgr.addResult( request.getClientID().intValue(), request.getReqID().intValue(), tmpDescription);
		                
		            } else {
		                //
		                // TODO - here do not treat it for the case of remote request ?
		                //
		                // case of a remote result
		                ObjectRemoteRequest object = new ObjectRemoteRequest();
					    object.setRequestId( request.getReqID().intValue() );
					    object.setDescription( tmpDescription );
					    object.setRequest( request );
					    object.setType( Object4Queue.REMOTE_REQUEST4RESULT );
					    blockingQueue.add( object );
		            }
		        }
			}
		}
		return found;
	}
	
	
	/**
	 * Checks if the lifetime is still valid or not.
	 * 
	 * @param description
	 * @return true if the lifetime is valid, false otherwise.
	 */
	private boolean checkLifetime( MSDDescription description ) {
	    
	    CreationInformation creation = description.getCreaInfo();
	    int lifetime = creation.getLifetime().intValue();
	    if ( lifetime <= 0 )
	        creation.setLifetime( DEFAULT_LIFETIME );
	    
	    Calendar time = creation.getUpdTime();
	    
	    int updateNumber = creation.getUpdNb().intValue();
	    if ( updateNumber == 0 ) {
		    // first time the lifetime is check
	        time = Calendar.getInstance();
		    creation.setCreaTime( time );
		    updateNumber++;
		    creation.setUpdNb( BigInteger.valueOf( updateNumber ) );
		    return true;
	    } 
	    
	    // check if time + lifetime >= currentTime
	    Calendar currentTime = Calendar.getInstance();
	    Calendar descTime = time;
	    descTime.add( Calendar.SECOND, lifetime );
	    if ( descTime.before( currentTime ) ) 
	        return false;
	    
	    return true;
	}
	
	//
	// Needed for the GUI
	//
	
	/**
	 * Sets the ManagerGUI.
	 */
	public void setGUINotification( GUINotification gui ) {
	    this.gui = gui;
	}
	
	/**
	 * Gets the cache of local descriptions 
	 * 
	 * @return vector of MSDDescription
	 */
	public Vector getLocalDescriptions() {
	    return this.localMSDDescriptionList;
	}
	
	/**
	 * Gets the cache of remote descriptions
	 * 
	 * @return vector of MSDDescription
	 */
	public Vector getRemoteDescriptions() {
	    return this.remoteMSDDescriptionList;
	}
}
