/*
 * Created on May 18, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 



package org.msda.components.manager.task;

import java.util.Iterator;
import java.util.Vector;

import org.msda.components.interfaces.PluginInterface;
import org.msda.components.manager.MSDManagerPool;
import org.msda.components.manager.ThreadDescriptionMgr;
import org.msda.components.manager.object.Object4Queue;
import org.msda.components.manager.object.ObjectPlugin;
import org.msda.components.manager.object.ObjectRemoteRequest;
import org.msda.framework.FrameworkOutput;
import org.msda.msd.MSDDescription;
import org.msda.msd.MSDRequest;
import org.msda.util.InfoType;

/**
 * Task that manages the interactions with one plugin.<br>
 *    - discovery request <br>
 *    - (remote) service access <br>
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 */

public class TaskPlugin extends TaskAbstract {
    
   
    private MSDManagerPool msdManagerPool;
    private ThreadDescriptionMgr descManager;
    private PluginInterface plugin;
    private FrameworkOutput board;
    private String response;
    private ObjectPlugin object;
    
    /**
     * TaskPlugin.
     * 
     * @param plugin
     * @param msdManagerPool
     */
    public TaskPlugin( PluginInterface plugin, MSDManagerPool msdManagerPool ) 
    {
        this.msdManagerPool = msdManagerPool;
        this.descManager = msdManagerPool.getDescManager();
        this.plugin = plugin;
		this.board = org.msda.framework.FrameworkMain.msdaOutput;
        this.response = null;
        this.object = null;
    }
    
    /**
     * In the case of a (remote) service access.
     * 
     * @return Returns the SOAP response.
     */
    public String getResponse() {
        String tmp = response;
        this.response = null;
        return tmp;
    }
    
    
    /**
     * @see org.msda.components.manager.task.TaskAbstract#setObject(org.msda.components.manager.Object4Queue)
     */
    public void setObject(Object4Queue object) {
        this.object = (ObjectPlugin)object;
    }
    
    /**
     * @see org.msda.components.manager.task.TaskAbstract#terminate()
     */
    public void terminate() {
        
        if ( object.getRequest()==null )
	        // notify the TaskServiceAccess it has ended
	        synchronized( object ) {
	            object.notify();
	        }
        
        this.msdManagerPool = null;
        this.descManager = null;
        this.plugin = null;
        this.board = null;
        this.response = null;
        this.object = null;
    }
    
    /**
     * @see org.msda.components.manager.task.TaskAbstract#run()
     */
    public void run() {
    	long searchTime, start, end, timeout, currentTime, t1;
       
        MSDRequest request = object.getRequest();
        
        if ( request!=null ) 
        {
            //
            // DISCOVERY REQUEST
            //
        	board.debug( "Start a search with the plugin "+plugin.getName() );

            start = System.currentTimeMillis();
            timeout = request.getProcInfo().getCrea().getLifetime().intValue() * 1000;
            
            board.debug( "Timeout ="+timeout +" Start="+start );
            
            Vector vTmpRemote = null;

            end = 0;
            while( end < timeout ) {
                
//                if ( InfoType.TEST_PARTIAL_RESULTS && plugin.getName().equals("UPnP")) {
//                    board.info("skip UPnP");
//                    return;
//                }
                
                // Search of services in the plugin
                searchTime = System.currentTimeMillis();
                Vector serviceList = plugin.evaluateRequest( request );
                currentTime = System.currentTimeMillis();
                end = currentTime - start;
                searchTime = currentTime - searchTime;
                board.debug( "start "+ start +" search start "+searchTime+" search end "+currentTime );
                board.debug( "TaskPlugin: End a search with the plugin "+plugin.getName()+" after "+end );
                board.debug( "TaskPlugin: Found "+serviceList.size()+" results...");
            
                // If the search time is too high, there is no need to search again with the plugin
                t1 = searchTime+searchTime;
                if ( t1 > timeout ) 
                    end = timeout;
                
                MSDDescription tmpDescr;
                for (int y=0; y<serviceList.size(); y++)
                {
                    tmpDescr = (MSDDescription)serviceList.get(y);
                    
                    if ( !object.isRemote() ) {
                        // if the object(request) is a local object, then we look for the
                        // corresponding pending request and add the results there. this will 
                        // in turn wake up the thread for sending back the msd description to the client.
                        int clientId = request.getClientID().intValue();
                        int requestId = request.getReqID().intValue();

                        boolean added = msdManagerPool.getMSDManager().addResult( clientId, requestId, tmpDescr );
                        // adding the result to the cache of local description
                        // if it has been added to the PendingRequest to avoid redundant description
                        // NOTE : it is ok because the request lifetime << cached description lifetime
                        if ( added ) {
                            board.debug( "TaskPlugin : RESULT added to pending request " + clientId + " " + requestId);
                            descManager.addService( tmpDescr, true );
                        }
                        else {
                            board.debug( "TaskPlugin : WARNING: could not add a result in a pending request" + clientId + " " + requestId);
                        }
                        
                    } else {
                    	// if the object is a remote request, then we put the result into an object to be
                    	// send back to the source network
                        if ( vTmpRemote==null )
                            vTmpRemote = new Vector();
                        
                        // As the plugin was in passive mode, we add the result to the cache of local description
                        // only if it has not been already added
                        if ( addDescription( tmpDescr, vTmpRemote ) )
                            descManager.addService( tmpDescr, true );
                        
                        // create a remote request object and add the result in it
                        // then put it into the manager pool of objects to process
                        ObjectRemoteRequest objectRR = new ObjectRemoteRequest();
                        objectRR.setRequestId( request.getReqID().intValue() );
                        objectRR.setDescription( tmpDescr );
                        objectRR.setRequest( request );
                        objectRR.setType( Object4Queue.REMOTE_REQUEST4RESULT );
                        
                        // unuseful, performs the task as soon as possible...
                        //msdManagerPool.getBlockingQueue().add( objectRR );
                        TaskAbstract task = new TaskRemoteRequest4Result( msdManagerPool );
    					task.setObject( objectRR );
    					task.run();
                    }
                }

                // if TEST (only first result returned we exit here...
                if ( InfoType.TEST_PARTIAL_RESULTS ) 
                {
                    board.debug("TaskPlugin : exit with the first result...");
                	break;
                }
            }
            vTmpRemote = null;

        } else {
            //
            // SERVICE ACCESS (remote or local)
            //
            
            // Service access from the plugin to the service
            if ( object.getSoapMessage()==null || object.getEpService()==null ) {
                
                board.info( "TaskPlugin::run : Object is not correct..."+object.getSoapMessage()+" "+object.getEpService() );
                
            } else {
                board.debug( "TaskPlugin::run : plugin.sendToService()..." );
                //board.info( "TaskPlugin::run : plugin.sendToService()..." );
                start = System.currentTimeMillis(); 
                response =  plugin.sendToService(object.getSoapMessage(), object.getEpService() ) ;
                end = System.currentTimeMillis() - start; 
                
                //board.info( "TaskPlugin::run : response after..."+end + " "+response );
                board.debug( "TaskPlugin::run : the response is :"+response+" after "+end);
                board.debug( "TaskPlugin::run : "+object.getSoapMessage());
                board.debug( "TaskPlugin::run : "+object.getEpService());
            }
        }
    }
    
    /**
     * Adds a description to a temporary collection. 
     * 
     * @param desc	the MSD description
     * @param v		the collection of MSD descriptions 
     * @return true if the MSD description has been added, false otherwise
     */
    private boolean addDescription( MSDDescription desc, Vector v ) {
        
        for (Iterator iter = v.iterator(); iter.hasNext();) {
            MSDDescription element = (MSDDescription) iter.next();
            boolean condition = element.getDescID().equals( desc.getDescID() );
            if ( condition ) 
                break;
            
            if ( !condition ) {
                v.add( desc );
                return true;
            }
        }
        
        return false;
    }
    
    //
	// Needed for the GUI
	//
    
    /**
     * @return Returns the plugin.
     */
    public PluginInterface getPlugin() {
        return plugin;
    }
}
