/*
 * Created on June 28, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.plugins;

import org.msda.components.interfaces.*;
import org.msda.components.plugins.http.HTTPListener;
import org.msda.components.plugins.http.HTTPServer;
import org.msda.components.plugins.http.httpHolder;
import org.msda.components.plugins.msdaservice.MSDService;
import org.msda.msd.*;
import org.msda.msd.impl.MSDRequestImpl;
import org.msda.util.*;

import org.msda.soap.SOAPRequest;
import org.msda.soap.SOAPTools;

import http.msg.HttpMsg;

import java.io.StringReader;

import javax.xml.namespace.QName;


/**
 * @author Rafik CHIBOUT
 */
public abstract class AbstractPlugin implements PluginInterface, HTTPListener {
	
	protected String creator="";
	protected String name="";
	protected Info info;
	protected String[][] parameters = new String[20][2];
	protected int paramPosition = 0;
	
	
	protected TranslatorInterface translator;
	protected PluginCallback callbckInterface;
	protected RequestInterface msdInterface;
	
	protected boolean activate = false;
	
	protected int httpServerPort = 8088;
	protected int directMSDPort = 7777;
	
	private MSDService service = new MSDService();
	private SOAPTools soapTools = null;
	
	private static final String SVC_ACCESS = "/discovery/serviceaccess";
	
	
	public AbstractPlugin()
	{
		try
		{
			soapTools = new SOAPTools("configuration/MSDDeploy.wsdd");
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
		
	}
	
	/**
	 * Returns the plugin name.
	 */
	public String getName()
	{
		return info.getProtocolName();
	}
	
	
	
	/**
	 * implements the http listener which catchs the HTTP message sent to the plugin and performs it.
	 * 
	 * @see org.msda.components.plugins.http.HTTPListener#newHttpMessage(org.msda.plugins.http.httpHolder)
	 */
	synchronized public void newHttpMessage(httpHolder messageHolder)
	{
		//board.info("AbstractPlugin : the plugin received an http message");
		HttpMsg message = null ;
		if (messageHolder != null ) message = messageHolder.getHTTPMessage();
		if (message != null)
		{
			String content = new String(message.getContent());
			String serviceName = message.getFirstLineToken(1);
			
			if (serviceName.equals("/discovery")) {
				// SOAP message contains a discovery request (MSDRequest) or client ID request 
				StringReader reader = new StringReader(content);
				SOAPRequest soapRequest = (SOAPRequest)soapTools.SOAP2Java(reader, "MSDClientAPI");
				
				if (soapRequest.getOperation().equals(QName.valueOf("{urn:MSDClientAPIService}getServices")))
				{
					// SOAP messages contains a discovery request
					MSDRequestImpl msdRequest = (MSDRequestImpl)soapRequest.getParametersValue()[0];
					MSDDescription[] msdDescs = service.getServices(msdRequest);
					String soapResponse = soapTools.SOAPResponse(msdDescs, new Object[0], "MSDClientAPI", QName.valueOf("{urn:MSDClientAPIService}getServices"));
					//send soapResponse to client
					try
					{
						HTTPServer.sendHttpResponse(messageHolder.getOutputStream(),soapResponse.getBytes());
					}	
					catch(Exception ex)
					{
						ex.printStackTrace();
					}
					return;
				}
			}
			else if (serviceName.indexOf("/discovery/serviceaccess") == 0) {
				//board.info("AbstractPlugin : the plugin received a Services Access Request " + serviceName );
				// SOAP message contains a Services Access Request
				// MSDaddress/epservice/msdDescId
				String epService = serviceName.substring( SVC_ACCESS.length()+1 );
				int index = epService.lastIndexOf("/");
				String endpointService = epService.substring( 0, index);
				String msdDescId = epService.substring( index+1 );
				
				//board.info(" - epService = " + epService );
				//board.info(" - endpointService = "+endpointService);
				//board.info(" - msdDescId = "+msdDescId);
				
				// service access
				String soapResponse = service.serviceAccess(content, msdDescId, endpointService);
				// send the response to the client
				try
				{
					HTTPServer.sendHttpResponse(messageHolder.getOutputStream(), soapResponse.getBytes());
				}
				catch(Exception ex)
				{
					ex.printStackTrace();
				}
				return;
			}
		}
		
	}
	/*
	 * this method allows user to put the translator class of this plugin, which will be used to translate the MSDRequest
	 * to the request-format of the protocol discovery 
	 */
	
	public void setTranslator(String className)
	{
		Class translatorClass;
		try
		{
			translatorClass = Class.forName(className);
		}
		catch(ClassNotFoundException ex)
		{
			//logger.error("Class "+className+" not found");
			return;
		}
		try
		{
			translator = (TranslatorInterface)translatorClass.newInstance();
		}
		catch(InstantiationException ex)
		{
			//logger.error("at "+translatorClass.getName()+" instantiation");
		}
		catch(IllegalAccessException ex)
		{
			//logger.error("Illegal access to "+translatorClass.getName());
		}
	}
	
	/* PluginInterface interface methods */ 
	
	/*
	 * this method allows the plugin's creator to put the plugin's information, the httpServerPort will be used by the http Server created inside plugin   
	 */
	
	public void setPluginInfo(String name, String creator, int httpServerPort, int directMSDPort)
	{
		this.name=name;
		this.creator=creator;
		this.httpServerPort = httpServerPort;
		this.directMSDPort = directMSDPort;
	}
	
	/*
	 * this method complete the setPluginInfo method, by providing an open discription of plugin's parameters
	 * 
	 */
	
	public void setPluginParam(String name, String value) 
	{
		if (paramPosition<parameters.length)
		{
			parameters[paramPosition][0]=name;
			parameters[paramPosition][1]=value;
			paramPosition++;
		}
	}
	
	/*
	 *  here, the user have to give the plugin a reference of msdManager that will be used by the plugin to search and to access services 
	 *  
	 */
	public void registerManager(PluginCallback callbckInterface, RequestInterface msdInterface) throws MSDException
	{
		this.callbckInterface = callbckInterface;
		this.msdInterface    = msdInterface;
		service = new MSDService(msdInterface);
	}
	
	public void registerTransformer(String transformerClassName)
	{
		
	}
	
	
	
	/*
	 * this method activate the plugin: start the HTTP server, have been compelited by tje activation of the protocol discovery and others 
	 * things needed for the plugin running    
	 */	
	public void activate()
	{
		HTTPServer server = new HTTPServer();
		server.setListener(this);
		server.setHTTPPORT(httpServerPort);
		server.start();
	}
	
	
	/*
	 * getPluginInfo returns the plugin's information on string format
	 */
	
	public String getPluginInfo()
	{
		StringBuffer buf = new StringBuffer();
		buf = buf.append( "PluginName = " ).append( info.getName() ).append( System.getProperty( "line.separator" ) )
		.append( "Creator = " ).append( creator ).append( System.getProperty( "line.separator" ) )
		.append( "Protocol = " ).append( info.getProtocolName()).append( System.getProperty( "line.separator" ) )
		.append( "HTTPServer Port = ").append( httpServerPort ).append( System.getProperty( "line.separator" ) )
		.append( "DirectMSD Port = " ).append( directMSDPort ).append( System.getProperty( "line.separator" ) );
		return buf.toString();
	}
	
	
	/*
	 * getInfo returns the plugin's information
	 */
	public Info getInfo()
	{
		return info;
	}
	
	/*
	 * this method send the soap message (service access message) to the service, and returns a soap message response of the service invocation.
	 * 
	 */
	
	public String sendToService(String sOAPMessage, String serviceEP)
	{
		return ServiceInvocation.invokService(sOAPMessage, serviceEP);
	}
	
	
}
