/*
 * Created on May 3, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.components.utils;

import java.util.List;
import java.util.Vector;

import org.msda.framework.FrameworkMain;
import org.msda.framework.FrameworkOutput;
import org.msda.msd.InformationComponent;
import org.msda.msd.NetworkConfig;
import org.msda.util.MSDUtil;

/**
 * This class is used to keep and manage the local components available on the network
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 */
public class LocalComponentsMgr {

    private Vector cacheOfManagers;
    private Vector cacheOfDevices;
    private Vector cacheOfBridges;
    
    private FrameworkOutput board;
    
    public LocalComponentsMgr() {
        this.cacheOfManagers = new Vector();
        this.cacheOfBridges = new Vector();
        this.cacheOfDevices = new Vector();
    }
    
    public void setBoard( FrameworkOutput board ) {
        this.board = board;
    }
    
    /**
     * @return Returns the cacheOfBridges.
     */
    public Vector getBridges() {
        return cacheOfBridges;
    }
    /**
     * @return Returns the cacheOfDevices.
     */
    public Vector getDevices() {
        return cacheOfDevices;
    }
    /**
     * @return Returns the cacheOfManagers.
     */
    public Vector getManagers() {
        return cacheOfManagers;
    }
    
    public NetworkConfig getNetworkConfig( InformationComponent info, String networkName ) {
    	NetworkConfig netConfig;
        for( int i=0; i<info.getActvNets().size(); i++) {
            netConfig = (NetworkConfig)info.getActvNets().get(i);
            if ( netConfig.getName().equals( networkName ) )
                return netConfig;
        }
        return null;
    }
    
    /**
     * Gets the bridge InformationComponent from two network names
     * 
     * @param srcNetworkName	the source network name
     * @param destNetworkName	the destination network name
     * @return InformationComponent
     */
    public InformationComponent getBridge( String srcNetworkName, String destNetworkName ) {
     
        InformationComponent infoTmp = find( srcNetworkName, destNetworkName, cacheOfBridges );
        //ADLC TMP - what to do if more than 1 bridge
        //         - not optimized, first found
        return infoTmp;
    }
    
    /**
     * Gets the manager InformationComponent from its network name
     * 
     * @param networkName	the manager network name
     * @return InformationComponent
     */
    public InformationComponent getManager( String networkName ) {
    	InformationComponent info;

    	for( int i=0; i<cacheOfManagers.size(); i++) {
    		info = (InformationComponent)cacheOfManagers.get(i);
            if ( info.getNetName().equals( networkName ) )
            	return info;
    	}
        return null;
    }
    

    
    /**
     * Adds a new Manager to the cache of Managers or updates it
     * 
     * @param info	the new Manager
     * @return true if the manager has been added, false otherwise
     */
    public boolean addManager ( InformationComponent info  ) {
    	if ( !update( info, cacheOfManagers ) ) {
    		cacheOfManagers.add( info );
	        FrameworkMain.msdaGUI.addManager();
    		return true;
    	}
    	return false;
    }

    /**
     * Adds a new Bridge to the cache of Bridges or updates it
     * 
     * @param info the new Bridge
     * @return true if the bridge has been added, false otherwise
     */
    public boolean addBridge ( InformationComponent info ) {
        if ( !update( info, cacheOfBridges ) ) {
        	cacheOfBridges.add( info );
	        FrameworkMain.msdaGUI.addBridge();
        	return true;
        }
        return false;
    }

    /**
     * Adds a new Component to the cache of Components or updates it
     * 
     * @param info	the new Component
     * @return
     */
    public boolean addComponent ( InformationComponent info  ) {
        if ( !update( info, cacheOfDevices ) ) {
            cacheOfDevices.add( info );
            return true;
        }
        return false;
    }

    /**
     * Removes a saved Manager based on its name
     * 
     * @param name	its name
     * @return	true if it was removed, false otherwise
     */
    public boolean removeManager( InformationComponent removedInfo ) {
    	InformationComponent info;

    	for (int i=0; i<cacheOfManagers.size(); i++) {
    		info = (InformationComponent)cacheOfManagers.get(i);
    		if ( info.getCompID().getName().equals( removedInfo.getCompID().getName() ) ) {
    			cacheOfManagers.remove( i );
       			FrameworkMain.msdaGUI.removeManager(removedInfo.getCompID().getName());
    			return true;
    		}
    	}
    	return false;
    }
    
    /**
     * Removes a saved Bridge based on its name 
     * 
     * @param name	its name
     * @return	true if it was removed, false otherwise
     */
    public boolean removeBridge( InformationComponent removedInfo ) {
    	InformationComponent info;
    	for (int i=0; i<cacheOfBridges.size(); i++) {
    		info = (InformationComponent)cacheOfBridges.get(i);
    		if ( info.getCompID().getName().equals( removedInfo.getCompID().getName() ) ) {
    			cacheOfBridges.remove( i );
    			FrameworkMain.msdaGUI.removeBridge(removedInfo.getCompID().getName());
    			return true;
    		}
    	}
    	return false;
    }
    
    /**
     * Removes a saved Component based on its name
     * 
     * @param id 	its name
     * @return	true if it was removed, false otherwise
     */
    public boolean removeComponent( InformationComponent removedInfo ) {
    	InformationComponent info;
    	for (int i=0; i<cacheOfDevices.size(); i++) {
    		info = (InformationComponent)cacheOfDevices.get(i);
    		if ( info.getCompID().getName().equals( removedInfo.getCompID().getName() ) ) {
    			cacheOfDevices.remove( i );
    			return true;
    		}
    	}
    	return false;
    }
    
    /**
     * Updates the saved InformationComponent if any in the vector
     * 
     * @param info	the new InformationComponent
     * @param v		the vector of saved InformationComponent
     * @return	true if something has been updated, false otherwise
     */
    private boolean update( InformationComponent info, Vector v ) {
    	InformationComponent infoTmp;
    	
        for( int i=0; i<v.size(); i++) {
            infoTmp = (InformationComponent)v.get(i);
            if ( //infoTmp.getNetworkName().equals( info.getNetworkName() ) && 
                 MSDUtil.getInstance().compareID( infoTmp.getCompID(), info.getCompID() ) ) {
                
                v.setElementAt( info, i );
                return true;
            }
        }
        return false;
    }
    
    /**
     * Finds if a network exists in a list of NetworkConfig.
     * 
     * @param networkName	The network name
     * @param networks		The list of NetworkConfig
     * @return NetworkConfig if the network Name exists in the list, null otherwise
     */
    private NetworkConfig find( String networkName, List networks ) {
    	NetworkConfig netConfig;
    	
        for ( int i=0;i<networks.size();i++) {
            netConfig = (NetworkConfig)networks.get(i);
            if ( networkName.equals( netConfig.getName() ) )
                return netConfig;
        }
        return null;
    }
    
    /**
     * Finds a saved MSD Stack that has two networks from a vector of InformationComponent 
     * 
     * @param srcNetworkName
     * @param destNetworkName
     * @param v
     * @return The corresponding InformationComponent, null otherwise
     */
    private InformationComponent find( String srcNetworkName, String destNetworkName, Vector v ) {
    	InformationComponent infoTmp;
    	List list;
    	NetworkConfig netConfig;
    	
        for( int i=0; i<v.size(); i++) {
            infoTmp = (InformationComponent)v.get(i);
            list = infoTmp.getActvNets();
            for( int j=0; j<list.size(); j++) {
                netConfig = (NetworkConfig)list.get(j);
                if ( netConfig.getName().equals( srcNetworkName ) ) {
                    for(int k=0; k<list.size();k++) {
                        if (k!=j) {
                            netConfig = (NetworkConfig)list.get(k);
                            if ( netConfig.getName().equals( destNetworkName ) ) 
                                return infoTmp;
                        }
                    }
                }
            }
        }
        return null;
    }
}
