/*
 * Created on May 11, 2005
 *
 * Comments, questions: 
 *	Pierre-Guillaume.Raverdy@inria.fr
 *	Rafik.Chibout@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.gui;

import java.awt.*; 
import java.util.Vector;

import org.msda.components.manager.MSDManager;
import org.msda.gui.manager.MenuDisplay;
import org.msda.gui.manager.SplitPanel;
import org.msda.gui.manager.ToolbarPanel;
import org.msda.msd.DeviceConfig;

/**
 * GUI that shows the MSD Manager :<br>
 * 	- Select "Plugins" to see the plugins <br>
 *  - Select "Local descriptions" to see the local cached descriptions <br>
 *  - Select "Remote descriptions" to see the remote cached descriptions <br>
 *  - Select "Bridges" to see the bridges<br>
 *  - Select "Statistics" to see some statistical information <br>
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 *
 */
public class ManagerGUI extends Panel implements GUINotification, GUIMessageNotification {
    
    public static final Color COLOR = new Color( 230,230,230 );
    public static final Color SPECIAL = new Color( 207,237, 222 );
    public static final Color NEGATIVE_COLOR =  Color.BLACK;
    
    public static final int WIDTH = 210;
    public static final int HEIGHT = 210;
    
    private SplitPanel splitPanel;
    private ToolbarPanel toolbarPanel;
    private MenuDisplay menuDisplay;
    
    private DeviceConfig config;
    private MSDManager manager;
    
    public ManagerGUI( MSDManager manager, DeviceConfig config ) { 
        
        setSize( WIDTH, HEIGHT ); 
        setBackground( ManagerGUI.COLOR );
        
        this.manager = manager;
        this.config = config;
        
        toolbarPanel = new ToolbarPanel( this );
        toolbarPanel.setBackground( SPECIAL );
        splitPanel = new SplitPanel();
        splitPanel.setBackground( SPECIAL );
        
        toolbarPanel.setSize( ManagerGUI.WIDTH, 60 );
        splitPanel.setSize( ManagerGUI.WIDTH, ManagerGUI.HEIGHT-80 );
        
        menuDisplay = new MenuDisplay( this );
        
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints c = new GridBagConstraints();
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.fill = GridBagConstraints.BOTH;
        gridbag.setConstraints( toolbarPanel, c);
        gridbag.setConstraints( splitPanel, c);
        
        setLayout( gridbag );
        add( toolbarPanel );
        add( splitPanel );
    } 
    
    /**
     * Returns the menuDisplay.
     * 
     * @return Returns the menuDisplay.
     */
    public Menu getMenuDisplay() {
        return menuDisplay;
    }
    
    /**
     * Get the device characteristics
     *  
     * @return Vector of (DeviceConfig, ... )
     */
    public Vector getDevice() {
        Vector v = new Vector();
        if ( config!=null )
            v.add( config );
        return v;
    }
    
    /**
     * Gets the list of the plugins. 
     * 
     * @return Vector of ( PluginInterface, ... ) 
     */
    public Vector getPlugins() {
        Vector res = manager.getPlugins();
        if ( res==null )
        	return new Vector();
        return res;
    }
    
    /**
     * Gets the list of the local cached descriptions 
     * 
     * @return Vector of (MSDDescription, ...)
     */
    public Vector getLocalDescriptions() {
        return manager.getLocalDescriptions();
    }
    
    /**
     * Gets the list of the remote cached descriptions 
     * 
     * @return Vector of (MSDDescription, ...)
     */
    public Vector getRemoteDescriptions() {
        return manager.getRemoteDescriptions();
    }
    
    /**
     * Gets the list of the active remote accesses 
     * 
     * @return
     */
    public Vector getBridges() {
        
        if ( manager!=null && manager.getLocalMgr()!=null )
            return manager.getLocalMgr().getBridges();
        return new Vector();
    }
        
    /**
     * Returns the SplitPanel
     * 
     * @return Returns the SplitPanel
     */
    public SplitPanel getSplitPanel() {
        return splitPanel;
    }
    
    /**
     * Returns the toolbarPanel.
     * 
     * @return Returns the toolbarPanel.
     */
    public ToolbarPanel getToolbarPanel() {
        return toolbarPanel;
    }
    
    //
    // Notification
    //
    
    public void addBridge() {
        if ( MenuDisplay.BRIDGES.equals( menuDisplay.getActivePanel() ) ) {
            menuDisplay.actionBridges();
        }
    }
    
    public void removeBridge(String name) {
        if ( MenuDisplay.BRIDGES.equals( menuDisplay.getActivePanel() ) ) {
            menuDisplay.actionBridges();
        }
    }

    public void addManager() {
        // nothing to do
    }

    public void removeManager(String name) {
        // nothing to do
    }

    /**
     * Notify the GUI that a description has been added.
     * 
     * @param local true if the description is a local description, false otherwise
     */
    public void addDescription( boolean local ) {
        menuDisplay.setToNullDescriptions( local );
        
        // This is the current panel, updates it
        if ( MenuDisplay.DESCRIPTIONS.equals( menuDisplay.getActivePanel() ) 
             || MenuDisplay.DESCRIPTIONS_REMOTE.equals( menuDisplay.getActivePanel() ) ) {
            if ( local )
                menuDisplay.actionDescriptionsLocal();
            else 
                menuDisplay.actionDescriptionsRemote();
        }
    }
    
    /**
     * Notify the GUI that a description has been removed.
     * 
     * @param local	true if the description is a remote description, false otherwise
     */
    public void removeDescription( boolean local ) {
        
        menuDisplay.setToNullDescriptions( local );
        
        // This is the current panel, updates it
        if ( MenuDisplay.DESCRIPTIONS.equals( menuDisplay.getActivePanel() ) 
             || MenuDisplay.DESCRIPTIONS_REMOTE.equals( menuDisplay.getActivePanel() ) ) {
            if ( local )
                menuDisplay.actionDescriptionsLocal();
            else
                menuDisplay.actionDescriptionsRemote();
        }
    }
    
    
    /**
     * @see org.msda.gui.GUIMessageNotification#notify(java.lang.Object, java.lang.String)
     */
    public void notify(Object o, String from) {
        // TODO Auto-generated method stub

    }
    /**
     * @see org.msda.gui.GUIMessageNotification#notify(java.lang.Object)
     */
    public void notify(Object o) {
        // TODO Auto-generated method stub

    }
    /**
     * @see org.msda.gui.GUIMessageNotification#notifyError(java.lang.Object)
     */
    public void notifyError(Object error) {
        // TODO Auto-generated method stub

    }
} 

