/*
 * Created on Apr 19, 2005
 *
 * @author raverdy
 * Comments, questions: Pierre-Guillaume.Raverdy@inria.fr
 *
 * INRIA, ARLES Project
 * 2004-2005
 */ 


package org.msda.util;

import java.util.List;
import java.util.Vector;

import org.msda.context.ContextMatcher;
import org.msda.msd.*;

/**
 * Utility class to compare msd data. 
 * <p>
 * Usage : MSDUtil.getInstance()
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 */
public class MSDUtil {

	
    private static final MSDUtil instance = new MSDUtil();
    
    // private constructor to prevent from creating instances of this class
    private MSDUtil() {}
    
    /**
     * Returns a static instance of MSDUtil 
     * 
     * @return Returns a static instance of MSDUtil 
     */
    public static MSDUtil getInstance() {
        return MSDUtil.instance;
    }
    
    /**
     * Verifies if a Vector contains a description with the same name than an input description.
     * <p>
     * NOTE : verify also the creator manager id to avoid to reject the same service in two different networks  
     * 
     * @param v		the Vector of MSDDescription
     * @param entry	the input description
     * @return true if the Vector contains a description with the same name than the entry
     */
    public boolean containsDescription( Vector v, MSDDescription entry ) {
        
        if ( v==null || v.isEmpty() )
            return false;
        
        String name = entry.getSvcDef().getGeneric().getName();
        String mgrID = entry.getCreaInfo().getCreator().getMgrID().getName();
        
        MSDDescription tmpDesc;
        for( int i=0; i<v.size(); i++) {
            tmpDesc = (MSDDescription)v.get(i);
            if ( name.equals( tmpDesc.getSvcDef().getGeneric().getName() ) 
                 && mgrID.equals( tmpDesc.getCreaInfo().getCreator().getMgrID().getName() ) )
                return true;
        }
        
        return false;
    }
    
    
    /**
	 * Compares two UbisecIDs
	 * 
	 * @param id1 ubisec id to compare
	 * @param id2 ubisec id to compare
	 * @return true if the two IDs are the same, false otherwise
	 */
	public boolean compareID (UbisecID id1, UbisecID id2) {
		
		if (id1.getName()!=null && id2.getName()!=null)
			if (!id1.getName().equals(id2.getName()))
				return false;

		return true;
	}
    
    
    /**
	 * Compares a msd description with another msd description.<br>
	 *    - compares the creators<br>
	 *    - compares the service definition <br>
	 * TODO - to see others criteria (propagationInformation, ...) ?
	 * 
	 * @param filter 
	 * @param entry 
	 * @return true is msd description matches, false otherwise
	 */
	public boolean matchingDescription ( MSDDescription filter, MSDDescription entry )
	{
	    //board.info( "_____________matchingDescription_______________" );
	    
	    // compares the creators
	    CreationInformation ciFilter = filter.getCreaInfo();
	    CreationInformation ciEntry = entry.getCreaInfo();
	    
	    if ( ciFilter!=null && ciEntry!=null ) {
	        
	        if ( ciFilter.getCreator()!=null ) {
	            if ( !ciFilter.getCreator().equals( ciEntry.getCreator() ) )
	                return false;
	        }
	    }
	    
	    // compares the service definition
	    ServiceDefinition sdEntry = entry.getSvcDef();
	    ServiceDefinition sdFilter = filter.getSvcDef();
	    if ( sdFilter!=null && sdEntry!=null ) {
	        
	        Generic gFilter = sdFilter.getGeneric();
	        Generic gEntry = sdEntry.getGeneric();
	        if ( gFilter!=null && gEntry!=null ) {
	            
	           // board.info( "AGNES filter.getName()="+gFilter.getName()+" and "+gEntry.getName() );
	            if ( !gFilter.getName().equals( gEntry.getName() ) )
	                return false;
	            
//	            board.info( "filter.getAddress()="+gFilter.getAddress()+" and "+gEntry.getAddress() );
//		        
//	            if ( !gFilter.getAddress().equals( gEntry.getAddress() ) )
//	                return false;
//	            
//	            board.info( "filter.getBinding()="+gFilter.getBinding()+" and "+gEntry.getBinding() );
//		        
//	            if ( !gFilter.getBinding().equals( gEntry.getBinding() ) )
//	                return false;
//	            
//	            board.info( "filter.getName()="+gFilter.getName()+" and "+gEntry.getName() );
//		         
//	            if ( !gFilter.getName().equals( gEntry.getName() ) )
//	                return false;
//	            
//	            // compares the attributes list
//	            if ( gFilter.getParams()!=null && gEntry.getParams()!=null ) {
//		            
//	                board.info( "filter.getParams()="+gFilter.getParams().size()+" and "+gEntry.getParams().size() );
//	 		       
//	                if ( gFilter.getParams().size()!=gEntry.getParams().size() )
//	                    return false;
//	                    
//		            Iterator it = gFilter.getParams().iterator();
//		            while( it.hasNext() ) {
//		                Param att = (Param)it.next();
//		                if ( !gEntry.getParams().contains( att ) )
//		                    return false;
//		            }
//		        }
//	            
//	            // compares the methods list
//	            if ( gFilter.getMethods()!=null && gEntry.getMethods()!=null) {
//	                
//	                board.info( "filter.getMethods()="+gFilter.getMethods().size()+" and "+gEntry.getMethods().size() );
//		 		       
//	                if ( gFilter.getMethods().size()!=gEntry.getMethods().size() )
//	                    return false;
//	                
//	                Iterator it = gFilter.getMethods().iterator();
//	                while( it.hasNext() ) {
//			            Method method = (Method)it.next();
//			            if ( !gEntry.getMethods().contains( method ) )
//			                return false;
//			        }
//	            }
	        }
	        
	        return true;
	    }
		return false;
	}
	
	
	/**
	 * Compares a partial service from a discovery request with a cached msd description.<br>
	 * This method is used when processing a msd request to look for relevant msd description.
	 * <p>
	 * 
	 * @param filter 	The partial service description
	 * @param entry 	The description
	 * @param rules		Context rules
	 * @return true if the partial service description matches the description
	 */
	public boolean matchingRequest ( RequestService filter, MSDDescription entry, MSDRequest request )
	{
	    // NOTE ADLC : trying to match only the name with "indexOf"
	    
	    //board.info( "________matchingRequest_______" );
	    
	    if ( request!=null && request.getCtxRules()!=null ) {
	        
	        if ( entry.getCtxInfo()!=null ) {
	            ContextMatcher matcher1 = new ContextMatcher(request, entry.getCtxInfo().getCISvc() );
	            if ( matcher1.matchUserService()==false )
	                return false;
	        }
	    }
	    
	    ServiceDefinition sd = entry.getSvcDef();
	    if (sd!=null) {
	        Generic generic = sd.getGeneric();
//	        board.info( "filter.getAddress()="+filter.getAddress()+" and "+generic.getAddress() );
//	        if ( filter.getAddress()!=null ) {
//	            if ( !filter.getAddress().equals( generic.getAddress() ) )
//	                return false;
//			}
//	        board.info( "filter.getBinding()="+filter.getBinding()+" and "+generic.getBinding() );
//	        if ( filter.getBinding()!=null ) {
//	            if ( !filter.getBinding().equals( generic.getBinding() ) )
//	                return false;
//	        }
	        //board.info( "filter.getName()="+filter.getName()+" and "+generic.getName() );
	        if ( filter.getName()!=null ) {
	            if ( generic.getName().indexOf( filter.getName() ) < 0) 
	            	return false;
	        }
	        //board.info( "return true");
	        return true;
//	        board.info( "filter.getAttributes()="+filter.getParams());
//	        if ( filter.getParams()!=null ) {
//	           
//	            Iterator it = filter.getParams().iterator();
//	            while( it.hasNext() ) {
//	            	Param att = (Param)it.next();
//	                if ( !generic.getParams().contains( att ) )
//	                    return false;
//	            }
//	        }
//   
//	        if ( filter.getMethods()!=null ) {
//	             
//	            Iterator it = filter.getMethods().iterator();
//		        while( it.hasNext() ) {
//		            Method method = (Method)it.next();
//		            
//		            boolean isOK = false;
//		            for ( int i=0;i<generic.getMethods().size();i++) {
//		                Method methodGeneric = (Method)generic.getMethods().get(i);
//		                if ( methodGeneric.getName().equals( method.getName() ) ) {
//		                    // compares the inputs
//			                if ( !method.getInputs().isEmpty() ) {
//			                    isOK = compareMethod( method.getInputs(), methodGeneric.getInputs() );
//			                    if ( !isOK )
//			                        return false;
//			                }
//			                // compares the outputs
//			                if ( !method.getOutputs().isEmpty() ) {
//			                    isOK = compareMethod( method.getOutputs(), methodGeneric.getOutputs() );
//			                    if ( !isOK )
//			                        return false;
//			                }
//			                isOK = true;
//			                break;
//		                }
//		            }
//		            if ( !isOK )
//		                return false;
//		        }
//	        }
	        
	    
	    }
		return false;
	}
	
	/**
	 * Compares the two list of Param.
	 * 
	 * @param l1
	 * @param l2
	 * @return true if every Param of the list  l1 has been found in the list l2, false otherwise
	 */
	private boolean compareMethod( List l1, List l2 ) {
	    
	    // compares the name and type of outputs
	    // if type != ""
        for( int j=0;j<l1.size();j++) {
            Param p = (Param)l1.get(j);
            //board.info( "compareMethod for param "+p.getType());
            int k=0;
            for( ; k<l2.size(); k++) {
                Param pGeneric= (Param)l2.get(k);
                if ( pGeneric.getType().equals( p.getType() ) ) {
                    if ( !p.getValue().equals("") ) {
                        if ( pGeneric.getValue().equals( p.getValue() ) ) {
                            break;
                        }
                    } else {
                        // compare only the name, it's ok
                        break;
                    }
                }
            }
            if ( k==l2.size() ) {
                // could not find the method
                return false;
            }
        }
        // every param of l1 has been found in l2
        return true;
	}
}
