/*
 * Created on Apr 1, 2005
 * 
 * UPC
 * 
 */

package org.msda.util.security;


import java.security.InvalidKeyException;
import java.security.Key;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import javax.crypto.*;
import javax.crypto.spec.*;

import secureMulticast.event.*;

/**
 * This class is in charge of signing and encoding Strings and bytes arrays
 * depending on the network the entity is going to be commnicated.
 * 
 * Class modified by Juan (UPC) to integrate the LKH algorithm.
 * 
 * TODO: Maybe we should drop definitely the support of Strings and use
 * only byte array. Study as well the possibility of using a not write
 * encoded byte array for signing.
 * 
 */
public class SecurityUtil implements LKHListener{
    
	/** A hastable of instances of this class */
	private static java.util.Hashtable instances=new java.util.Hashtable(); 
	/**
	 * Name of the digest algorithm 
	 * TODO - in the source code and by default... to change as a parameter ?
	 */
	private static final String DIGEST_ALGORITHM = "MD5";
	private static final String ENCRYPTION_ALGORITHM = "DES";
	private static final String MAC_ALGORITHM = "HmacSHA1";
	private static final byte[] KEY_SEED = {1,2,3,4,5,6,7,8};
	
	private Key key;
	private MessageDigest hasher;
	private Cipher encrypter=null;
	private Cipher decrypter=null;
	private Mac mac;
	
	private Object keySync = new Object();
	private Object encrypterSync = new Object();
	private Object decrypterSync = new Object();
	private Object macSync = new Object();
	

	/** Creates the security util with a default (and well known!) key.
	 * Keep in mind this class will not be really secured until an LKHEvent
	 * jumps.
	 */
	public SecurityUtil() {
		try {
			setGroupKey(new SecretKeySpec(KEY_SEED,ENCRYPTION_ALGORITHM));
			
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}		
	}
	
	/** @return The key used to sign the communication */
	public Key getGroupKey()
	{
		synchronized (keySync) {
			return key;
		}
	}
	
	/** Sets the group key of this class.
	 * @param key The key
	 * @throws NoSuchAlgorithmException
	 * @throws NoSuchPaddingException
	 * @throws InvalidKeyException
	 */
	public void setGroupKey(Key key) throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException
	{
		synchronized (keySync) {
			this.key = key;
			hasher = MessageDigest.getInstance(DIGEST_ALGORITHM);
			synchronized (encrypterSync) {
				encrypter = Cipher.getInstance(this.key.getAlgorithm());
				encrypter.init(Cipher.ENCRYPT_MODE,key);
			}
			synchronized (decrypterSync) {
				decrypter= Cipher.getInstance(this.key.getAlgorithm());
				decrypter.init(Cipher.DECRYPT_MODE,key);
			}
			
			mac = Mac.getInstance(MAC_ALGORITHM);
			synchronized (macSync) {
				mac.init(key);
			}
			
			System.out.println("Group key is:"+transformToString(this.key.getEncoded()));
		}
	}
	
	
	/** Gets an instance for this class using an identifier.
	 * The multicast group or the networ interface can be used as
	 * identifier of the instance.
	 * 
	 * This method is useful to crete one and only one SecurityUtil for
	 * eacj network.
	 * 
	 * @param id The identifier used.
	 * @return The SecurityUtil object.
	 */
	public static SecurityUtil getInstance(String id) {
		SecurityUtil s=(SecurityUtil)instances.get(id);
		if(s==null){
			return createInstance(id);
		}else{
			return s;
		}
	}
	
	/** Creates an instance of this class for an identifier.
	 * @param id The identifier of this instance. The multicast address
	 * or the name of the network interface can be used as identifier. 
	 */
	private static SecurityUtil createInstance(String id){
		SecurityUtil s=new SecurityUtil();
		instances.put(id,s);
		return s;
	}
	
	/**
	 * Creates the checksum corresponding to a String using the algorithm
	 * @param text 
	 * @return The binary checksum of the input text
	 */
	public synchronized byte[] createChecksum( byte[] message ) {
		synchronized (hasher) {
			hasher.reset();
			return hasher.digest(); 
		}
	}
	
	/** @param text A string 
	 * @return A printable version of the cheksum from a String */
	public String createChecksum( String text ) {
		return transformToString( createChecksum(text.getBytes()));            
	}
	
	/** @param message A byte array to create a MAC
	 * @return The binary mac of this byte array
	 */
	public byte[] createMac(byte[] message) {
		synchronized (macSync) {
			return mac.doFinal(message);
		}
	}
	
	/** @param text A String to create a MAC.
	 * @return A printable version of the MAC of the string.
	 */
	public String createMac(String text) {
		return transformToString(createMac(text.getBytes()));
	}
	
	/** Encrypts a message.
	 * 
	 * @param message Message to be encrypted
	 * @return A binary encryption of the message.
	 * @throws IllegalBlockSizeException
	 * @throws BadPaddingException
	 */
	public byte[] encrypt(byte[] message) throws IllegalBlockSizeException, BadPaddingException{
		synchronized (encrypterSync) {
			return encrypter.doFinal(message);
		}
	}
	
	/** Decrypts a message.
	 * 
	 * @param message The byte array to be decrypted.
	 * @return A binary representation of the message
	 * @throws IllegalBlockSizeException
	 * @throws BadPaddingException
	 */
	public byte[] decrypt(byte[] message) throws IllegalBlockSizeException, BadPaddingException{
		synchronized (decrypterSync) {
			return decrypter.doFinal(message);
		}

	}
	
	/** Transfrom an array of bytes (binary form) into a printable,
	 * human readable and XML storable String.
	 * @param byteText A binary byte array 
	 * @return A printable version of the byte array
	 */
	protected String transformToString( byte[] byteText ) {
		
		StringBuffer hexString = new StringBuffer();
		for (int i=0;i<byteText.length;i++) {
			hexString.append ( Integer.toHexString(0xFF & byteText[i]) );
		}
		
		return hexString.toString();
	}
	
	/** Gets an event from the LKH library. We are
	 * interested in the REKEING events.
	 * @param e The event.
	 */
	public void LKHEventPerformed(LKHEvent e) {
		try {
			switch(e.getEvent()){
			case LKHEvent.JOINING:
			case LKHEvent.LEAVING:
				((secureMulticast.server.LKHserver)e.getSource()).getAlgorithm().process();
				break;
			case LKHEvent.NEW_SEK:
				setGroupKey(e.getNewSEK());
			}
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}
	
	/** @return The length in bytes of the checksum. */
	public int getChecksumLength(){
		return hasher.getDigestLength();
	}
	
	/** @return The length in bytes of the mac */
	public int getMacLength(){
		return mac.getMacLength();
	}
}
