/*
 * Created on Aug. 18, 2005
 *
 * INRIA, ARLES Project
 * 
 * contact: Pierre-Guillaume.Raverdy@inria.fr<br>
 *          Rafik.Chibout@inria.fr
 * 
 * Copyright INRIA 2005
 */
package org.msda.clientsamples.upnp;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import org.msda.MSDClientAPIService.MSDClientAPIBindingStub;

import com.siemens.upnp.InvalidDescriptionException;
import com.siemens.upnp.UPnPConstants;
import com.siemens.upnp.controlpoint.ControlPoint;
import com.siemens.upnp.controlpoint.ProxyDevice;
import com.siemens.upnp.controlpoint.ProxyService;
import com.siemens.upnp.controlpoint.UPnPSearchResponseListener;

/**
 * Abstract client to defines a Client adapted to a UPnP service.
 * 
 * @author agnes.de_la_chapelle@inria.fr - INRIA ARLES
 *
 */
public abstract class AbstractClientUPnP {

    //
    // Methods to search and access using UPnP
    //
    
    protected ControlPoint controlPoint;
    protected String deviceDescURL;
    protected String UPnPDesc;
    protected ProxyService upnpService;
    
    /**
     * Finds a UPnP service using its name
     * 
     * @param name	the service name
     * @return	the URL of the device description 
     */
    public String findUPnPService( String name ) {
        
        UPnPDesc = "";
        try {
            HttpURLConnection conn = getConnection( name );
            if ( conn==null )
                throw new IOException("Can't connect to the service");
            
            BufferedInputStream reader = new BufferedInputStream(conn.getInputStream());
            byte [] buff = new byte[1000];
            while (reader.available()>0)
            {
                int num= reader.read(buff,0,buff.length);
                UPnPDesc = UPnPDesc + (new String(buff,0,num));
            }
            //System.out.println("");
            //System.out.println("**************");
            
            //System.out.println("Level 1");
            URL url = conn.getURL();
            ProxyDevice device = controlPoint.createDevice(url);
            
            //System.out.println("Level 2: Device created");
            java.util.Vector services = device.getServices();
            
            //System.out.println("Level 3: service proxy get");
            if (services != null && services.size()>0)
            {
                upnpService = (ProxyService)services.elementAt(0);
                System.out.println("ServiceCtrURL="+upnpService.getControlURL());
                return deviceDescURL;
            }
            
        } catch (IOException e) {
            e.printStackTrace();
        } catch (InvalidDescriptionException e) {
            e.printStackTrace();
        }
        return null;
    }
    
    
    /**
     * @return Returns the deviceDescURL.
     */
    protected String getDeviceDescURL() {
        return deviceDescURL;
    }
   
    /**
     * @return Returns the uPnPDesc.
     */
    public String getUPnPDesc() {
        return UPnPDesc;
    }
    
    /**
     * @return Returns the upnpService.
     */
    public ProxyService getUpnpService() {
        return upnpService;
    }
    
    //
    // Methods to search and access the service using the MSD 
    //
    
    private static final String MSD_SERVICE_NAME = "msdservice:1:0";
    
    protected String ctrUrl4MSDService;
    protected org.msda.MSDClientAPIService.MSDClientAPIPortType msdService;
    
    
    /**
     * Finds the MSD service using UPnP
     */
    protected void findMSDService() {
        
        ctrUrl4MSDService = "";
        
        // search MSDService
        try {
            HttpURLConnection conn = getConnection( MSD_SERVICE_NAME );
            if ( conn==null )
                throw new IOException("Can't connect to the service");
            
            URL url = conn.getURL();
            
            BufferedInputStream reader = new BufferedInputStream(conn.getInputStream());
            byte[] buff = new byte[1000];
            while (reader.available()>0)
            {
                int num= reader.read(buff,0,buff.length);
                System.out.print(new String(buff,0,num));
            }
            //System.out.println("");
            //System.out.println("**************");
            
            //System.out.println("Level 1");
            ProxyDevice device = controlPoint.createDevice(url);
            
            //System.out.println("Level 2: Device created");
            java.util.Vector services = device.getServices();
            
            //System.out.println("Level 3: service proxy get");
            
            // Here, takes the first
            if (services != null && services.size()>0) {
                ProxyService service = (ProxyService)services.elementAt(0);
                ctrUrl4MSDService = service.getControlURL();
                
                System.out.println("MSDService Found at "+ctrUrl4MSDService);
                msdService = new MSDClientAPIBindingStub( ctrUrl4MSDService );
            }
            else
            {
                System.out.println("Cannot find MSDService");
            }
            
        } catch (IOException e) {
            e.printStackTrace();
        } catch (InvalidDescriptionException e) {
            e.printStackTrace();
        }
       
        controlPoint.shutDown();
    }
    
    //
    // Private methods...
    //
    
    /**
     * Gets the HTTP connection to a UPnP service
     * 
     * @param serviceType	The type of the service
     * @return the connection
     */
    private HttpURLConnection getConnection( String serviceType ) {
        
        controlPoint = ControlPoint.getInstance();
        DeviceDesc deviceDesc = new DeviceDesc();
        UPnPSearchResponseListener searchResponseListener = new UPnPSearchResponseListenerImpl(deviceDesc);
        controlPoint.addUPnPSearchListener(searchResponseListener);
        
        try {
            //System.out.println(": Level 1 ServiceType="+serviceType);	
            controlPoint.search(UPnPConstants.SearchForServiceType(serviceType));
            //System.out.println(": Level 2");
            while (deviceDesc.getDescriptionURL() == null ) {}
            //System.out.println("DescrURL = "+deviceDesc.getDescriptionURL());
            //System.out.println(": Level 3");
            deviceDescURL = deviceDesc.getDescriptionURL();
            URL url = new URL( deviceDescURL );
            //System.out.println(": Level 4");
            return (HttpURLConnection)url.openConnection();
            
        } catch (MalformedURLException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }
}
